/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include "all.h"
#include "SourceLine.h"
#include "gprint.h"

void polar_xy(double r, double angle, double *dx, double *dy);
void xy_polar(double dx,double dy,double *radius,double *angle);

int glefitcf_(int *mode,float *x, float *y, int *l, int *m, float *u, float *v, int *n);

void fitbez(double **pxv, double **pyv, int **pmv, int *pnp, bool multi) {
	int np = *pnp;
	if (np > 200 || np <= 2) {
		return;
	}
	vector<float> xin(np);
	vector<float> yin(np);
	for (int i = 0; i < np; i++) {
		xin[i] = (float)(*pxv)[i];
		yin[i] = (float)(*pyv)[i];
	}
	int mode = (multi ? 2 : 1);
	int nsub = 300/(np-1);
	if (nsub <= 1) nsub = 2;
	int np_out = (np-1)*nsub+1;
	vector<float> xout(np_out);
	vector<float> yout(np_out);
	glefitcf_(&mode, &xin[0], &yin[0], &np, &nsub, &xout[0], &yout[0], &np_out);
	double* xv = (double*)myallocz(np_out*sizeof(double));
	double* yv = (double*)myallocz(np_out*sizeof(double));
	int* m = (int*)myallocz(np_out*sizeof(int));
	for (int i = 0; i < np_out; i++) {
		xv[i] = (double)xout[i];
		yv[i] = (double)yout[i];
		m[i] = 0;
	}
	*pxv = xv;
	*pyv = yv;
	*pmv = m;
	*pnp = np_out;
}

void transform_log(double* v, int np) {
	for (int i = 0; i < np; i++) {
		v[i] = log10(v[i]);
	}
}

void untransform_log(double* v, int np) {
	for (int i = 0; i < np; i++) {
		v[i] = pow(10.0, v[i]);
	}
}

void fitbez_log(double **pxv, double **pyv, int **pmv, int *pnp, bool multi, bool islog) {
	if (islog) {
		int np = *pnp;
		vector<double> logv(np);
		for (int i = 0; i < np; i++) {
			logv[i] = log10((*pyv)[i]);
		}
		*pyv = (double*)&logv[0];
		fitbez(pxv, pyv, pmv, pnp, multi);
		untransform_log(*pyv, *pnp);
	} else {
		fitbez(pxv, pyv, pmv, pnp, multi);
	}
}
