/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#define GRAPHDEF extern

#include "all.h"
#include "tokens/stokenizer.h"
#include "SourceLine.h"
#include "bitmap/img2ps.h"
#include "op_def.h"
#include "mem_limits.h"
#include "token.h"
#include "gleinterface.h"
#include "glearray.h"
#include "polish.h"
#include "pass.h"
#include "var.h"
#include "mygraph.h"
#include "graph.h"
#include "begin.h"
#include "color.h"
#include "leastsq.h"
#include "core.h"
#include "axis.h"
#include "SourceLine.h"
#include "file_io.h"
#include "cutils.h"
#include "gprint.h"
#include "key.h"
#include "glearray.h"
#include "run.h"
#include "sub.h"
#include "numberformat.h"

#define DEFAULT_STEPS 100

extern int g_nbar;
extern axis_struct xx[];

extern vector<int> g_fcalls;
extern vector<int> g_funder;
extern GLEColorMap* g_colormap;

extern int g_nkd, g_keycol;

template <class T>
inline void maxminval(T* array,T* max_v, T* min_v, int size)
{

	register int 	i;

	*max_v = array[0];
 	*min_v = array[0];
//		printf("min=%0.5lf max=%0.5lf\n",*min_v,*max_v);
	for(i = 1 ; i < size ; i++ ){
		*max_v = (array[i] > *max_v)? array[i]:*max_v;
		*min_v = (array[i] < *min_v)? array[i]:*min_v;
//		printf("min=%0.5lf max=%0.5lf\n",*min_v,*max_v);
  	}
}

void box3d(double x1, double y1, double x2, double y2,double x3d,double y3d,int sidecolor, int topcolor, int notop);
void fitbez_log(double **pxv, double **pyv, int **pmv, int *pnp, bool multi, bool islog);
void var_clear_local(void);
void big_open(char *s) throw (ParserError);
void big_vec(int nomiss);
void windownorm(void);
void windowdn(int d);
void big_mark(int mnum, double msize) throw (ParserError);
void do_svg_smooth(double* , int);
void draw_herrbar(double x, double y, double eup, double ewid);
double fnx(double vv);

void do_draw_lines(double* xt, double* yt, int* m, int npts);
void do_draw_steps(double* xt, double* yt, int* m, int npts);
void do_draw_fsteps(double* xt, double* yt, int* m, int npts);
void do_draw_hist(double* xt, double* yt, int* m, int npts);
void do_draw_bar(double* xt, double* yt, int* m, int npts);
void do_draw_impulses(double* xt, double* yt, int* m, int npts);

double fnx(double vv) {
	if (data_negate[GLE_AXIS_X]) {
		vv = wxmax - (vv-wxmin);
	}
	if (xx[GLE_AXIS_X].log) {
		return (log10(vv)-log10(wxmin))/(log10(wxmax)-log10(wxmin)) * xlength + xbl;
	} else {
		return (vv-wxmin)/(wxmax-wxmin) * xlength + xbl;
	}
}

double fny(double vv) {
	if (data_negate[GLE_AXIS_Y]) {
		vv = wymax - (vv-wymin);
	}
	if (xx[GLE_AXIS_Y].log) {
		return (log10(vv)-log10(wymin))/(log10(wymax)-log10(wymin)) * ylength + ybl;
	} else {
		return (((vv-wymin)/(wymax-wymin)) * ylength + ybl);
	}
}

double graph_xgraph(double v) {
	// cout << "graph_xmin: " << graph_xmin << " graph_xmax: " << graph_xmax << endl;
	// cout << "graph_x1: " << graph_x1 << " graph_x2: " << graph_x2 << endl;

	if (graph_xmax == graph_xmin) return 0.0;
	if (data_negate[GLE_AXIS_X]) {
		v = (graph_xmax - v) + graph_xmin;
	}
	if (xx[GLE_AXIS_X].log) {
		return graph_x1 + (log10(v)-log10(graph_xmin))/(log10(graph_xmax)-log10(graph_xmin))*(graph_x2-graph_x1);
	} else {
		return graph_x1 + (v-graph_xmin)/(graph_xmax-graph_xmin)*(graph_x2-graph_x1);
	}
}

double graph_ygraph(double v) {
	if (graph_ymax == graph_ymin) return 0.0;
	if (data_negate[GLE_AXIS_Y]) {
		v = (graph_ymax - v) + graph_ymin;
	}
	if (xx[GLE_AXIS_Y].log) {
		return graph_y1 + (log10(v)-log10(graph_ymin))/(log10(graph_ymax)-log10(graph_ymin))*(graph_y2-graph_y1);
	} else {
		return graph_y1 + (v-graph_ymin)/(graph_ymax-graph_ymin)*(graph_y2-graph_y1);
	}
}

int bigalli;
static int bigcol1,bigcol2,bigally;
static int bigrecord;
FILE *fptr;

void big_open(char *infile2) throw(ParserError) {
	char infile[90];
	char *s1,*s2;
	strcpy(infile,infile2);
	if (infile[strlen(infile)-1] == '$') {
		int idx,typ;
		var_find(infile,&idx,&typ) ;
		if (idx>=0) var_getstr(idx,infile);
	}
	s1 = strchr(infile,',');
	bigrecord = 0;
	bigally = bigalli = 0;
	bigcol1 = 1; bigcol2 = 2;
	if (s1!=NULL) {
		s2 = strchr(s1+1,',');
		bigcol1 = atoi(s1+1);
		if (s2!=NULL) {
			if (*(s2+1) == '*') {
				bigally = true;
			} else {
				bigcol2 = atoi(s2+1);
				if (bigcol2==0) {
					gprint("Expecting \"file.name,xcoloumn,ycolumn\" found [%s] \n",infile);
				}
			}
			*s1 = 0;
		}
	}
	if (infile[strlen(infile)-1] == '$') {
		int idx,typ;
		var_find(infile,&idx,&typ) ;
		if (idx>=0) var_getstr(idx,infile);
	}
	validate_file_name(string(infile), true);
	fptr = fopen(infile,"r");
	if (s1!=NULL) *s1 = ',';
	if (fptr==NULL) {
		g_throw_parser_error_sys("unable to open data file '", infile, "'");
	}
}

void big_line(double *x1, double *y1, int *m) {
	static char *bigee;
	int bg1;
	static int ntk;
	if (fptr==NULL) return;
next_line:;
	if (bigally) {
		if (bigalli==0 || bigalli>=ntk) {
			if (bigee!=NULL) { myfree(bigee); bigee = 0;}
			if (feof(fptr)) return;
			fgets(ebuff,390,fptr);
			bigee = sdup(ebuff);
			bigalli = 0;
		}
		strcpy(ebuff,bigee);
		token_data(ebuff,tk,&ntk,tkbuff);
		*x1 = ++bigrecord;
		*y1 = atof(tk[++bigalli]);
		*m = false;
		if (feof(fptr)) {*m = true; return;}
		return;
	}
	if (feof(fptr)) return;
	fgets(ebuff,390,fptr);
	token_data(ebuff,tk,&ntk,tkbuff);
	if (ntk<bigcol1 || ntk<bigcol2) goto next_line;
	bg1 = bigcol1;
	if (bg1==0) bg1 = bigcol2;
	if (*tk[bg1] != '*' && *tk[bigcol2] != '*') {
		if (bigcol1==0) *x1 = ++bigrecord;
		else *x1 = atof(tk[bigcol1]);
		*y1 = atof(tk[bigcol2]);
		*m = false;
	} else { *m = true;}
	if (feof(fptr)) {*m = true; return;}
}

void setrange(double x, double y, int m) {
	if (m) return;
	if (x>range_x2) range_x2 = x;
	if (x<range_x1) range_x1 = x;
	if (y>range_y2) range_y2 = y;
	if (y<range_y1) range_y1 = y;
}

void setxrange(double x, int m) {
	if (m) return;
	if (x > range_x2) range_x2 = x;
	if (x < range_x1) range_x1 = x;
}

void setyrange(double y, int m) {
	if (m) return;
	if (y > range_y2) range_y2 = y;
	if (y < range_y1) range_y1 = y;
}

void preview_big(void) {
	int dn;
	double x1,y1;
	int m1;

	for (dn=1;dn<=ndata;dn++) {
	  if (dp[dn]!=NULL) {
	   if (dp[dn]->bigfile!=NULL && dp[dn]->autoscale) {
		big_open(dp[dn]->bigfile);
		if (fptr==NULL) goto endbigm;
		if (feof(fptr))  goto xxx;
		for (;;) {
			if (feof(fptr))  goto xxx;
			big_line(&x1,&y1,&m1);
			setrange(x1,y1,m1);
		}
xxx:		big_close();
endbigm:	;
	   }
	  }
	}
}

void request() {
	static char s[20];
	gprint("Press return to continue \n");
	fgets(s, 20, stdin);
}

double bar_get_min_interval(int bar, int idx = 0) {
	double min = 1e30;
	int d = br[bar]->to[idx];
	double* px = dp[d]->xv;
	for (int i = 1; i < dp[d]->np; i++) {
		double w = px[i] - px[i-1];
		if (w > 0 && w < min) min = w;
	}
	return min;
}

bool bar_has_type(bool horiz) {
	for (int bar = 1; bar <= g_nbar; bar++) {
		if (horiz == br[bar]->horiz) return true;
	}
	return false;
}

bar_struct::bar_struct() {
	ngrp = 0;
	width = 0.0;
	dist = 0.0;
	x3d = 0.0;
	y3d = 0.0;
	notop = false;
	horiz = false;
	for (int i = 0; i < 20; i++) {
		from[i] = 0;
		to[i] = 0;
		lwidth[i] = 0.0;
		lstyle[i][0] = 0;
		fill[i] = GLE_COLOR_BLACK;
		color[i] = GLE_COLOR_BLACK;
		side[i] = GLE_COLOR_BLACK;
		top[i] = GLE_COLOR_BLACK;
		pattern[i] = -1;
	}
}

void draw_bars() {
	int b,nb,d,i,di,df,dt;
	double *px,*py,min_int,*xt,*yt,*yf,bwid,whole_wid,bdis,yfval;
	int *m;

	for (b = 1; b <= g_nbar; b++) {
		done_line = true;
		if (br[b]==0) { gprint("Error, bars struct zero \n"); return;}
		nb = br[b]->ngrp;	/* number of bars in each group */
		d = br[b]->to[0]; 	/* dataset for first bar */
		if (d==0 || dp[d]==NULL)  {gprint("Error, bars zero dataset \n");return;}
		/* find the smallest x distance in dataset 1 */
		px = dp[d]->xv;
		py = dp[d]->yv;
		if (px==NULL || nb==0) { gprint("error in bar data dn=%d  ngrp=%d\n",d,nb); return;}
		m = dp[d]->miss;
		min_int = bar_get_min_interval(b);
		if (br[b]->width==0) br[b]->width = min_int/(nb*2);
		if (br[b]->dist==0) br[b]->dist = br[b]->width*1.4 ;
		bwid = br[b]->width;
		bdis = br[b]->dist;

		g_gsave();
		for (di = 0; di < nb; di++) {
			df = br[b]->from[di];
			dt = br[b]->to[di];
		  	whole_wid = (nb-1)*bdis+bwid;

			g_set_line_width(br[b]->lwidth[di]);
			g_set_line_style(&br[b]->lstyle[di][0]);
			if (br[b]->color[di] == 0) br[b]->color[di] = GLE_COLOR_BLACK;
			g_set_color(br[b]->color[di]);
			if (br[b]->pattern[di] != -1 && br[b]->pattern[di] != (int)GLE_FILL_CLEAR) {
				g_set_fill(br[b]->pattern[di]);
				g_set_pattern_color(br[b]->fill[di]);
			} else {
				g_set_fill(br[b]->fill[di]);
				g_set_pattern_color(GLE_COLOR_BLACK);
			}

			df = br[b]->from[di];
			dt = br[b]->to[di];
			if (dp[df]==NULL || dp[dt]==NULL) {
				gprint("No data in bargraph datasets\n");
				goto bar_exit;
			}
			yf = dp[df]->yv;
			xt = dp[dt]->xv;
			yt = dp[dt]->yv;
			m = dp[dt]->miss;
			if (yt==NULL) {
				gprint("No data in bargraph dataset. d(%d) \n",dt);
				goto bar_exit;
			}

			windowdn(d);
			for (i = 0;i < dp[d]->np; i++, xt++, yt++ ,m++) {
				if (yf != NULL) yfval = *yf++;
				else yfval = 0;
				if (!*m) {
					draw_bar((*xt)-whole_wid/2+di*bdis, yfval, *yt, bwid, br[b], di, b);
				}
			}
			windownorm();
		}
bar_exit:
		g_grestore();
	}
}

double graph_bar_pos(double xpos, int bar, int set) throw(ParserError) {
	if (set < 1 || set > g_nbar) {
		g_throw_parser_error("illegal bar set: ", set);
	}
	int nb = br[set]->ngrp;
	if (bar < 1 || bar > nb) {
		g_throw_parser_error("illegal bar number: ", bar);
	}
	double bwid = br[set]->width;
	double bdis = br[set]->dist;
	double whole_wid = (nb-1)*bdis+bwid;
	// cout << "bwid:  " << bwid << endl;
	// cout << "bdis:  " << bdis << endl;
	// cout << "nb:    " << nb << endl;
	// cout << "whole: " << whole_wid << endl;
	if (br[set]->horiz) {
		return graph_ygraph(xpos-whole_wid/2+(bar-1)*bdis+bwid/2);
	} else {
		return graph_xgraph(xpos-whole_wid/2+(bar-1)*bdis+bwid/2);
	}
}

void draw_bar(double x, double yf, double yt, double wd, bar_struct* barset, int di, int bar) throw(ParserError) {

	/* draw a bar, wd wide, centere at x , from yf, to yt */

	x = x + wd/2;
	double x1 = x - wd/2;
	double y1 = yf;
	double x2 = x + wd/2;
	double y2 = yt;

	double x3d = barset->x3d;
	double y3d = barset->y3d;
	int topcolor = barset->top[di];
	int sidecolor = barset->side[di];
	int notop = barset->notop;

	if (barset->horiz) {
		box_clip(&y1,&x1,wxmin,wymin,wxmax,wymax);
		box_clip(&y2,&x2,wxmin,wymin,wxmax,wymax);
		double bx1 = x1;
		double bx2 = x2;
		x1 = fnx(y1);  x2 = fnx(y2);
		y1 = fny(bx1); y2 = fny(bx2);
	} else {
		box_clip(&x1,&y1,wxmin,wymin,wxmax,wymax);
		box_clip(&x2,&y2,wxmin,wymin,wxmax,wymax);
		x1 = fnx(x1); x2 = fnx(x2);
		y1 = fny(y1); y2 = fny(y2);
	}


	if (x1 == x2 || y1 == y2) {
		// zero size bar
		return;
	}

	if (barset->style[di] == "") {
		if (x3d != 0.0) {
			box3d(x1, y1, x2, y2, x3d, y3d, sidecolor, topcolor, notop);
		}
		g_box_fill(x1, y1, x2, y2);
		g_box_stroke(x1, y1, x2, y2);
	} else {
		double args[7];
		args[0] = 0;
		args[1] = x1;
		args[2] = y1;
		args[3] = x2;
		args[4] = y2;
		args[5] = yt;
		args[6] = di;
		string name = string("BAR_") + barset->style[di];
		call_sub_byname(name, args, 6, "(used for defining bar style)");
	}
}

void box3d(double x1, double y1, double x2, double y2,double x3d,double y3d,int sidecolor, int topcolor, int notop) {
	/* assuming x3d is positive for the moment */
	double xx;

	if (x1>x2) { xx = x1; x1 = x2; x2 = xx;}
	if (y1>y2) { xx = y1; y1 = y2; y2 = xx;}

	x3d = x3d*(x2-x1);
	y3d = y3d*(x2-x1);

	if (x3d<0) {
		xx = x1; x1 = x2; x2 = xx;
	}
	g_gsave();
	g_set_path(true);
	g_set_line_join(1);	/* use rounded lines to avoid ucky peeks */
	g_newpath();
	g_move(x2,y1);
	g_line(x2+x3d,y1+y3d);
	g_line(x2+x3d,y2+y3d);
	g_line(x2,y2);
	g_line(x2,y1);
	if (topcolor!=0) {
	   g_set_fill(sidecolor);
	   g_fill();
	}
	g_stroke();
	g_newpath();

	if (!notop) {	/* now draw top of bar  */
	 g_move(x2,y2);
	 g_line(x2+x3d,y2+y3d);
	 g_line(x1+x3d,y2+y3d);
	 g_line(x1,y2);
	 g_line(x2,y2);
	 if (topcolor!=0) {
		g_set_fill(topcolor);
		g_fill();
	 }
	 g_stroke();
	}
	g_newpath();
	g_set_path(false);
	g_newpath();
	g_grestore();
}

void draw_user_function_calls(bool underneath) throw(ParserError) {
	vector<int>& calls = underneath ? g_funder : g_fcalls;
	if (calls.size() != 0) {
		string str;
		GLEParser* parser = get_global_parser();
		g_beginclip();
		g_set_path(true);
		g_newpath();
		g_box_stroke(xbl, ybl, xbl+xlength, ybl+ylength, false);
		g_clip();
		g_set_path(false);
		for (vector<int>::size_type i = 0; i < calls.size(); i++) {
			double res;
			int pln = calls[i];
			if (begin_line(&pln, str)) {
				parser->setString(str.c_str());
				Tokenizer* tokens = parser->getTokens();
				tokens->is_next_token_i("UNDER");
				GLEPcodeList pc_list;
				GLEPcode pcode(&pc_list);
				parser->get_subroutine_call(pcode);
				eval_pcode(pcode, &res);
			} else {
				g_throw_parser_error("unexpected empty line in graph block");
			}
		}
		g_endclip();
	}
}

void gr_thrownomiss(void) {
	for (int i=1;i<=ndata;i++) {
		if (dp[i]!=NULL) if (dp[i]->nomiss && dp[i]->np>0) {
			gr_nomiss(i);	/* throw away missing points */
		}
	}
}

void gr_nomiss(int i) {
	double *nx,*ny,*xt,*yt;
	int *m,*nm;
	int k,npnts;
	if (dp[i] == NULL) return;
	//cout << "gr nomiss" << i << endl;
	if ( dp[i]->xv!=NULL && dp[i]->yv!=NULL ) {
		k = 0;
		yt = dp[i]->yv;
		xt = dp[i]->xv;
		m =  dp[i]->miss;
		nx = xt;
		ny = yt;
		nm = m;
		npnts = dp[i]->np;
		for (int j=0 ; j<npnts ; j++,m++,xt++,yt++) {
			if (!*m) {
				*nx++ = *xt;
				*ny++ = *yt;
				*nm++ = *m;
				k++;
			}
		}
		dp[i]->np = k;
	}
}

void gr_nonan(int i) {
	if (dp[i] == NULL || dp[i]->xv == NULL || dp[i]->yv == NULL) return;
	double* yt    = dp[i]->yv;
	double* xt    = dp[i]->xv;
	int* m        = dp[i]->miss;
	int npnts     = dp[i]->np;
	int pos       = 0;
	bool ylog     = xx[2].log;
	bool xlog     = xx[1].log;
	for (int j = 0; j < npnts; j++) {
		bool ok = true;
		if (gle_isnan(xt[j])) ok = false;
		if (gle_isnan(yt[j])) ok = false;
		if (xlog && xt[j] < 0) ok = false;
		if (ylog && yt[j] < 0) ok = false;
		if (ok) {
			xt[pos] = xt[j];
			yt[pos] = yt[j];
			m[pos] = m[j];
			pos++;
		}
	}
	dp[i]->np = pos;
}

void roundrange(double *gmin, double *gmax, bool extend, bool tozero) {
	if (*gmax < *gmin) {
		*gmax = 100; *gmin = 10;
		return;
	}
	double delta = *gmax - *gmin;
	if (delta == 0) return;
	double st = delta/10;
	double expnt = floor(log10(st));
	double n = st/pow(10.0,expnt);
	int ni = 1;
	if (n > 5) {
		ni = 10;
	} else if (n > 2) {
		ni = 5;
	} else if (n > 1) {
		ni = 2;
	} else {
		ni = 1;
	}
	// auto-extend axis towards zero
	if (tozero && *gmin > 0 && *gmin <= delta*0.2) {
		*gmin = 0.0;
	}
	if (tozero && *gmax < 0 && *gmax >= -delta*0.2) {
		*gmax = 0.0;
	}
	// round range towards multiple of dticks
	double dticks = ni * pow(10.0,expnt);
	if (*gmin ==  floor( *gmin/ dticks) * dticks) {
		if (extend && *gmin != 0.0) *gmin = *gmin - dticks;
	} else {
		*gmin = (floor(*gmin/ dticks) * dticks );
	}
	if (( floor( *gmax/ dticks) * dticks) != (*gmax / dticks) * dticks ) {
		*gmax = floor(*gmax / dticks ) * dticks + dticks;
	} else {
		if (extend && *gmax != 0.0) *gmax = *gmax + dticks;
	}
}

bool window_min_max_set() {
	for (int i = 1; i <= 2; i++) {
		if (!xx[i].minset) return false;
		if (!xx[i].maxset) return false;
	}
	return true;
}

void window_set(bool makeup) {
	wxmin = range_x1;
	wxmax = range_x2;
	wymin = range_y1;
	wymax = range_y2;

	// cout << "xrange: " << wxmin << " - " << wxmax << endl;
	// cout << "yrange: " << wymin << " - " << wymax << endl;

	if (!xx[1].log) {
		bool has_horiz_bar = bar_has_type(true);
		roundrange(&wxmin, &wxmax, has_horiz_bar, false);
	}
	if (!xx[2].log) {
		bool has_vert_bar = bar_has_type(false);
		roundrange(&wymin, &wymax, has_vert_bar, true);
	}

	// cout << "xrange(r): " << wxmin << " - " << wxmax << endl;
	// cout << "yrange(r): " << wymin << " - " << wymax << endl;

	if (xx[GLE_AXIS_X].minset) wxmin = xx[GLE_AXIS_X].min;
	xx[GLE_AXIS_X].min = wxmin;
	if (!xx[GLE_AXIS_X2].minset) xx[GLE_AXIS_X2].min = wxmin;
	if (!xx[GLE_AXIS_X0].minset) xx[GLE_AXIS_X0].min = wxmin;

	if (xx[GLE_AXIS_X].maxset) wxmax = xx[GLE_AXIS_X].max;
	xx[GLE_AXIS_X].max = wxmax;
	if (!xx[GLE_AXIS_X2].maxset) xx[GLE_AXIS_X2].max = wxmax;
	if (!xx[GLE_AXIS_X0].maxset) xx[GLE_AXIS_X0].max = wxmax;

	if (xx[GLE_AXIS_Y].minset) wymin = xx[GLE_AXIS_Y].min;
	xx[GLE_AXIS_Y].min = wymin;
	if (!xx[GLE_AXIS_Y2].minset) xx[GLE_AXIS_Y2].min = wymin;
	if (!xx[GLE_AXIS_Y0].minset) xx[GLE_AXIS_Y0].min = wymin;

	if (xx[GLE_AXIS_Y].maxset) wymax = xx[GLE_AXIS_Y].max;
	xx[GLE_AXIS_Y].max = wymax;
	if (!xx[GLE_AXIS_Y2].maxset) xx[GLE_AXIS_Y2].max = wymax;
	if (!xx[GLE_AXIS_Y0].maxset) xx[GLE_AXIS_Y0].max = wymax;

	/* should go thru datasets to find max,min */
	if (makeup && (wxmin>=wxmax || wymin>=wymax)) {
		if (!xx[1].log) {
			wxmin = 0;
			wxmax = 10;
		} else {
			wxmin = 1;
			wxmax = 1e3;
		}
		if (!xx[2].log) {
			wymin = 0;
			wymax = 10;
		} else {
			wymin = 1;
			wymax = 1e3;
		}
	}
}

void store_window_bounds_to_vars() {
	var_findadd_set("XGMIN", xx[1].min);
	var_findadd_set("XGMAX", xx[1].max);
	var_findadd_set("YGMIN", xx[2].min);
	var_findadd_set("YGMAX", xx[2].max);
	var_findadd_set("X2GMIN", xx[3].min);
	var_findadd_set("Y2GMAX", xx[3].max);
	var_findadd_set("Y2GMIN", xx[4].min);
	var_findadd_set("Y2GMAX", xx[4].max);
}

void draw_fills() {
	double x1, y1, x2 = 0, y2 = 0, ymx, lastx, lasty;
	double *ssxt = NULL, *ssyt = NULL, *xt, *yt;
	int dn, i, n, *m, npnts, *ssm = NULL, free_smoothed=0;
	struct fill_data *ff;

	/* set clipping region */

	for (n=1;n<=nfd;n++) {
	  if (fd[n]->type==0) return;
	  done_line = true;
	  ff = fd[n];
	  if (wxmin > ff->xmin) ff->xmin = wxmin;
	  if (wxmax < ff->xmax) ff->xmax = wxmax;
	  if (wymin > ff->ymin) ff->ymin = wymin;
	  if (wymax < ff->ymax) ff->ymax = wymax;

	g_beginclip();		/* saves current clipping */
	g_set_path(true);
	g_newpath();
	g_move(fnx(ff->xmin),fny(ff->ymin));
	g_box_stroke(fnx(ff->xmin),fny(ff->ymin),fnx(ff->xmax),fny(ff->ymax));
	g_clip();		/* sets current path to be clipping */

	  vector<double> fvec;
	  dn = ff->da;
	  if (dp[dn]==NULL) { gprint("No data in fill dataset at all \n");return;}
	  free_smoothed = false;
	  yt = dp[dn]->yv;
	  xt = dp[dn]->xv;
	  m = dp[dn]->miss;
	  npnts = dp[dn]->np;

		if (dp[dn]->smooth && npnts>3 && npnts<190) {
			gr_nomiss(dn);	/* throw away missing points */
			npnts = dp[dn]->np;
			bool islog = xx[GLE_AXIS_Y].log;
			fitbez_log(&xt, &yt, &m, &npnts, dp[dn]->smoothm, islog);
			ssxt = xt; ssyt = yt; ssm = m;
			free_smoothed = true;
		}

	  ymx = ff->ymax;
	  if (xt==NULL) { gprint("No data in fill dataset \n"); return;}
	  x1 = *xt;
	  y1 = *yt;
	  switch(ff->type) {
	    case 1: /* x1,d1 */
		ymx = ff->ymin;
	    case 2: /* d1,x2 */
	        gr_nomiss(dn); /* throw away missing points */
		fill_vec(*xt,ymx,*xt,*yt,&fvec);
		for (i=0;i<npnts-1;i++,xt++,yt++) {
			fill_vec(*xt,*yt,*(xt+1),*(yt+1),&fvec);
		}
		fill_vec(*xt,*yt,*xt,ymx,&fvec);
		fill_vec(*xt,ymx,*dp[dn]->xv,ymx,&fvec);
		break;
	    case 3: /* d1,d2 */
		for (i=0;i<npnts-1;i++,m++,xt++,yt++) {
			fill_vec(*xt,*yt,*(xt+1),*(yt+1),&fvec);
			x2 = *(xt+1); y2 = *(yt+1);
		}
		dn = ff->db;
		yt = dp[dn]->yv;
		xt = dp[dn]->xv;
		m = dp[dn]->miss;
		npnts = dp[dn]->np;

		if (free_smoothed) {myfrees(ssxt,"Fill1"); myfrees(ssyt,"x"); myfrees(ssm,"y");}
		free_smoothed = false;
		if (dp[dn]->smooth && npnts>3 && npnts<190) {
			gr_nomiss(dn);	/* throw away missing points */
			npnts = dp[dn]->np;
			bool islog = xx[GLE_AXIS_Y].log;
			fitbez_log(&xt, &yt, &m, &npnts, dp[dn]->smoothm, islog);
			ssxt = xt; ssyt = yt; ssm = m;
			free_smoothed = true;
		}

		xt += npnts - 1;
		yt += npnts - 1;
		fill_vec(x2,y2,*xt,*yt,&fvec);
		for (i=0;i<npnts-1;i++,m--,xt--,yt--)
			fill_vec(*xt,*yt,*(xt-1),*(yt-1),&fvec);
		fill_vec(*xt,*yt,x1,y1,&fvec);
		break;
	    case 4: /* d1 */
		for (i=0;i<npnts-1;i++,m++,xt++,yt++) {
			if (!*m && !*(m+1))
				fill_vec(*xt,*yt,*(xt+1),*(yt+1),&fvec);
		}
		fill_vec(*xt,*yt,x1,y1,&fvec);
		break;
	  }

	if (free_smoothed) {myfrees(ssxt,"Fill2"); myfrees(ssyt,"f4"); myfrees(ssm,"f5");}
	/* Paint the region defined */
	g_set_fill(ff->color);
	g_newpath();
	if (fvec.size() > 2) {
		g_move(fnx(fvec[0]),fny(fvec[1]));
		lastx = fvec[0]; lasty = fvec[1];
		for (i = 0; i <= (int)fvec.size()-4; i += 4) {
			if (lastx!=fvec[i] || lasty!=fvec[i+1])  {
				g_closepath();
				g_move(fnx(fvec[i]),fny(fvec[i+1]));
			}
			g_line(fnx(fvec[i+2]),fny(fvec[i+3]));
			lastx = fvec[i+2]; lasty = fvec[i+3];
		}
	}
	g_closepath();
	g_fill();
	g_set_path(false);
	g_endclip();
	}
}

void fill_vec(double x1, double y1, double x2, double y2, vector<double>* vec) {
	vec->push_back(x1);
	vec->push_back(y1);
	vec->push_back(x2);
	vec->push_back(y2);
}

void setupdown(char *s, int *doup, int *upd, int *upp, double *upval) {
	static char vbuf[9];
	int j=0;

	*upd = 0;
	*doup = true;
	*upp = false;
	if (*s==0) { *doup=false; return;}
	if (toupper(*s)=='D') { *upd = atoi(s+1); return; }
	if (str_i_str(s,"%")!=NULL) { j=1; *upp = true; }
	strncpy(vbuf,s,strlen(s)-j+1);
	*upval = atof(vbuf);
}

/*------------------------------------------------------------------*/
/* 	d3 err .1			*/
/* 	d3 err 10%			*/
/* 	d3 errup 10% errdown d2		*/
/* 	d3 err d1 errwidth .2		*/
bool dataset_null(int i) {
	if (dp[i]==NULL) {
		gprint("Dataset %d doesn't exist at all\n",i);
		return true;
	}
	if (dp[i]->yv == NULL) {
		gprint("Dataset %d doesn't exist\n",i);
		return true;
	}
	return false;
}

void draw_herr(void) {
	int doup,upd,upp;
	double upval,eup,edown,ewid;
	int dodown,downd,downp;
	double downval;
	double *xt,*yt,x;
	int *m,mup,mdown;
	int dn,i;

	g_gsave();
	for (dn=1;dn<=ndata;dn++) {
	  dpp = dp[dn];
	  if (dpp!=NULL) if (dpp->herrup[0]!=0 || dpp->herrdown[0]!=0) {
		done_line = true;
		g_get_hei(&x);
		if (dpp->herrwidth==0) dpp->herrwidth = x/3;
		ewid = dpp->herrwidth;
		setupdown(dpp->herrup,&doup,&upd,&upp,&upval);
		setupdown(dpp->herrdown,&dodown,&downd,&downp,&downval);
		g_set_color(dpp->color);
		g_set_line_width(dpp->lwidth);
		windowdn(dn);
		yt = dpp->yv;
		xt = dpp->xv;
		m = dpp->miss;
		if (upd) if (dataset_null(upd)) return ;
		if (downd) if (dataset_null(downd)) return ;
		for (i=0;i<dpp->np;i++,m++,xt++,yt++) {
			mup = false; mdown = false;
			if (upd) {
				eup = *(dp[upd]->yv+i);
				mup = *(dp[upd]->miss+i);
			} else {
				eup = upval;
				if (upp) eup = (eup * *xt)/100;
			}
			if (downd) {
				edown = *(dp[downd]->yv+i);
				mdown = *(dp[downd]->miss+i);
			} else {
				edown = downval;
				if (downp) edown = (edown * *xt)/100;
			}
			/* if (eup!=-999) { */
			 if (doup) if (!*m) if (!mup) draw_herrbar(*xt,*yt,eup,ewid);
			 if (dodown) if (!*m) if (!mdown) draw_herrbar(*xt,*yt,-edown,ewid);
			/* } */
		}
		windownorm();
	  }
	}
	g_grestore();
}

void draw_herrbar(double x, double y, double eup, double ewid) {
	if (x>=wxmin && x<=wxmax && y>=wymin && y<=wymax ) {
		g_move(fnx(x),fny(y));
		g_line(fnx(x-eup),fny(y));
		g_move(fnx(x-eup),-ewid/2+fny(y));
		g_line(fnx(x-eup),ewid/2+fny(y));
	}
}

void draw_err() {
	int doup,upd,upp;
	double upval,eup,edown,ewid;
	int dodown,downd,downp,big;
	double downval;
	double *xt,*yt,x;
	int *m,mup,mdown;
	int dn,i = 0;

	g_gsave();
	for (dn=1;dn<=ndata;dn++) {
	  big = false;
	  dpp = dp[dn];
	  if (dpp!=NULL) if (dpp->errup[0]!=0 || dpp->errdown[0]!=0) {
		if (dpp->bigfile!=NULL) {
			big = true;
			big_open(dpp->bigfile);
	   	}
		done_line = true;
		g_get_hei(&x);
		if (dpp->errwidth==0) dpp->errwidth = x/3;
		ewid = dpp->errwidth;
		setupdown(dpp->errup,&doup,&upd,&upp,&upval);
		setupdown(dpp->errdown,&dodown,&downd,&downp,&downval);
		g_set_color(dpp->color);
		g_set_line_width(dpp->lwidth);
		windowdn(dn);
		yt = dpp->yv;
		xt = dpp->xv;
		m = dpp->miss;
		if (upd) if (dataset_null(upd)) return;
		if (downd) if (dataset_null(downd)) return;
		if (big) {{
		 	double x1,y1;
		 	int m2;

			if (fptr==NULL) goto endbigm;
			if (feof(fptr))  goto xxx;
			for (;;) {
aaa2:
				if (feof(fptr) )  goto xxx;
				big_line(&x1,&y1,&m2);
				if (m2) goto aaa2;
				mup = false; mdown = false;
				if (upd) {
					eup = *(dp[upd]->yv+i);
					mup = *(dp[upd]->miss+i);
				} else {
					eup = upval;
					if (upp) eup = (eup * y1)/100;
				}
				if (downd) {
					edown = *(dp[downd]->yv+i);
					mdown = *(dp[downd]->miss+i);
				} else {
					edown = downval;
					if (downp) edown = (edown * y1)/100;
				}
				if (doup) if (!mup) draw_errbar(x1,y1,eup,ewid);
				if (dodown) if (!mdown) draw_errbar(x1,y1,-edown,ewid);
			}
xxx:			fclose(fptr); fptr=NULL;
endbigm:		;
		}} else {
		  for (i=0;i<dpp->np;i++,m++,xt++,yt++) {
			mup = false; mdown = false;
			if (upd) {
				eup = *(dp[upd]->yv+i);
				mup = *(dp[upd]->miss+i);
			} else {
				eup = upval;
				if (upp) eup = (eup * *yt)/100;
			}
			if (downd) {
				edown = *(dp[downd]->yv+i);
				mdown = *(dp[downd]->miss+i);
			} else {
				edown = downval;
				if (downp) edown = (edown * *yt)/100;
			}
			/* if (eup!=-999) { */
			 if (doup) if (!*m) if (!mup) draw_errbar(*xt,*yt,eup,ewid);
			 if (dodown) if (!*m) if (!mdown) draw_errbar(*xt,*yt,-edown,ewid);
			/* } */
		  }
		}
		windownorm();
	  }
	}
	g_grestore();
	draw_herr();
}

void draw_errbar(double x, double y, double eup, double ewid) {
	if (x>=wxmin && x<=wxmax && y>=wymin && y<=wymax ) {
		g_move(fnx(x),fny(y));
		g_line(fnx(x),fny(y+eup));
		g_move(fnx(x)-ewid/2,fny(y+eup));
		g_line(fnx(x)+ewid/2,fny(y+eup));
	}
}

double dwx1,dwy1,dwx2,dwy2;

void windowdn(int d) {
	dwx1 = wxmin; dwy1 = wymin; dwx2 = wxmax; dwy2 = wymax;
	if (dp[d]==NULL) return;
	if (dp[d]->xmax > dp[d]->xmin) {
		wxmin = dp[d]->xmin; wxmax = dp[d]->xmax;
	}
	if (dp[d]->ymax > dp[d]->ymin) {
		wymin = dp[d]->ymin; wymax = dp[d]->ymax;
	}
}

void windownorm() {
	wxmin = dwx1 ; wymin = dwy1 ; wxmax = dwx2 ; wymax = dwy2 ;
}

void draw_lines() {
	double *ssxt = NULL,*ssyt = NULL, *xt, *yt, oldlwidth;
	int *m, *ssm = NULL;
	char oldlstyle[10];
	int dn,i,npnts;
	bool free_smoothed,big,free_deresolve;
	int j;
	g_gsave();
	g_get_line_style(oldlstyle);
	g_get_line_width(&oldlwidth);
	//printf("\nndata=%d\n",ndata);
	for (dn=1;dn<=ndata;dn++) {
	//printf("dn=%d",dn);
	  gr_nonan(dn);
	  big = false;
	  last_vecx = 1e10;
	  last_vecy = 1e10;
	  if (dp[dn]!=NULL)
	  {
	   if (dp[dn]->bigfile!=NULL)
	   {
		big=true;
		big_open(dp[dn]->bigfile);
	   }
	   if (dp[dn]->xv!=NULL || big)
	    if (dp[dn]->line==static_cast<int>(true) || dp[dn]->lstyle[0]!=0)
		{
/*		printf("lstyle=%d",dp[dn]->lstyle[0]);*/
		free_smoothed = false;
		g_set_line_style(oldlstyle); /* use defualts for each */
		g_set_line_width(oldlwidth); /* use defaults for each */
		g_set_line_style(dp[dn]->lstyle);
		g_set_color(dp[dn]->color);
		g_set_line_width(dp[dn]->lwidth);
		windowdn(dn);
		done_line = true;
		npnts = dp[dn]->np;
		if (dp[dn]->deresolve > 1) {
			/* throw away missing values */
			gr_nomiss(dn);
			/* compute resulting number of points and allocate arrays */
			npnts /= dp[dn]->deresolve;
			xt = (double*) myalloc(sizeof(double)*(npnts + 1));
			yt = (double*) myalloc(sizeof(double)*(npnts + 1));
			m =  (int*) myalloc(sizeof(int)*(npnts + 1));
			free_deresolve = true;
			if (!dp[dn]->deresolve_avg) {
				/* deresolve by just skipping some of the original points */
				for (i = 0, j = 0; i < npnts; i++, j += dp[dn]->deresolve) {
					xt[i] = dp[dn]->xv[j];
					yt[i] = dp[dn]->yv[j];
					m[i]  = false;
				}
			} else {
				/* deresolve by averaging points */
				/* d1 deresolve 5 average plots points something like */
				/* X = (X[5]-X[0]/2), Y = mean(Y[0]...Y[5]), etc. */
				for (i = 0; i < npnts; i++) {
					yt[i] = 0.0;
					for (j = 0; j < dp[dn]->deresolve; j++) {
						yt[i] += dp[dn]->yv[(i*dp[dn]->deresolve)+j];
					}
					yt[i] /= dp[dn]->deresolve;
					xt[i]  = (dp[dn]->xv[i*dp[dn]->deresolve]+
					          dp[dn]->xv[(i+1)*dp[dn]->deresolve-1])/2.0;
					m[i]   = false;
				}
			}
		} else {
			yt = dp[dn]->yv;
			xt = dp[dn]->xv;
			m = dp[dn]->miss;
			free_deresolve = false;
		}

		if (dp[dn]->smooth && npnts>3 && npnts<190)
		{
			/* (BUG, gr_nomiss has not been written HAS NOW !!! */
			gr_nomiss(dn);	/* throw away missing points */
			npnts = dp[dn]->np;
			bool islog = xx[GLE_AXIS_Y].log;
			fitbez_log(&xt, &yt, &m, &npnts, dp[dn]->smoothm, islog);
			ssxt = xt; ssyt = yt; ssm = m;
			free_smoothed = true;
			/* maybe use real bezier curves for drawing a smoothed graph ????? */
			/* instead of a huge number of line segments? */
		}

		if (dp[dn]->svg_smooth && npnts > 3 )
		{
//			cout <<"svg smooth";
			gr_nomiss(dn);	/* throw away missing points */
			if (dp[dn]->svg_iter == 0) dp[dn]->svg_iter =1;
			for(j=0;j<dp[dn]->svg_iter;j++)
			do_svg_smooth(dp[dn]->yv,dp[dn]->np);
		}

		if (big)
		{
/*			printf(" BIG npoints=%d ",npnts);*/
			big_vec(dp[dn]->nomiss);
		}
		else
		{
/*			printf(" npoints=%d ",npnts);*/
			switch (dp[dn]->line_mode) {
				case GLE_GRAPH_LM_PLAIN:
					do_draw_lines(xt, yt, m, npnts);
					break;
				case GLE_GRAPH_LM_STEPS:
					do_draw_steps(xt, yt, m, npnts);
					break;
				case GLE_GRAPH_LM_FSTEPS:
					do_draw_fsteps(xt, yt, m, npnts);
					break;
				case GLE_GRAPH_LM_HIST:
					do_draw_hist(xt, yt, m, npnts);
					break;
				case GLE_GRAPH_LM_IMPULSES:
					do_draw_impulses(xt, yt, m, npnts);
					break;
				case GLE_GRAPH_LM_BAR:
					do_draw_bar(xt, yt, m, npnts);
					break;
			}
		}
		if (free_deresolve) {
			myfree(xt);
			myfree(yt);
			myfree(m);
		}
		if (free_smoothed)
		{
			myfrees(ssxt,"Line1");
			myfrees(ssyt,"l2");
			myfrees(ssm,"l3");
		}
		windownorm();
	  }
	  big_close();
	 }
	 else
	 {
/*	printf(" NULL",dn);*/
	}/* end null*/
/*	printf("\n");*/
	}/* end for*/
	g_grestore();
}

void do_draw_lines(double* xt, double* yt, int* m, int npnts) {
	for (int i = 0; i < (npnts-1); i++, m++, xt++, yt++) {
		if (!*m && !*(m+1)) draw_vec(*xt,*yt,*(xt+1),*(yt+1));
	}
}

void do_draw_steps(double* xt, double* yt, int* m, int npnts) {
	// STEPS: (x1,y1) to (x2,y1) and the second from (x2,y1) to (x2,y2). See demo.
	for (int i = 0; i < (npnts-1); i++, m++, xt++, yt++) {
		if (!*m && !*(m+1)) {
			draw_vec(*xt,*yt,*(xt+1),*yt);
			draw_vec(*(xt+1),*yt,*(xt+1),*(yt+1));
		}
	}
}

void do_draw_fsteps(double* xt, double* yt, int* m, int npnts) {
	// FSTEPS: (x1,y1) to (x1,y2) and the second from (x1,y2) to (x2,y2)
	for (int i = 0; i < (npnts-1); i++, m++, xt++, yt++) {
		if (!*m && !*(m+1)) {
			draw_vec(*xt,*yt,*xt,*(yt+1));
			draw_vec(*xt,*(yt+1),*(xt+1),*(yt+1));
		}
	}
}

void do_draw_hist(double* xt, double* yt, int* m, int npnts) {
	double hist_prev_x = 0, hist_prev_y = 0, hist_x1, hist_x2;
	bool hist_has_prev = false;
	for (int i = 0; i < npnts; i++, m++, xt++, yt++) {
		if (!*m) {
			bool hist_valid = true;
			if (i < npnts-1 && !*(m+1)) {
				hist_x2 = (*(xt+1) + (*xt))/2;
				if (hist_has_prev) {
					hist_x1 = (hist_prev_x + (*xt))/2;
				} else {
					hist_x1 = 2*(*xt) - hist_x2;
				}
			} else {
				if (hist_has_prev) {
					hist_x1 = (hist_prev_x + (*xt))/2;
					hist_x2 = 2*(*xt) - hist_x1;
				} else {
					hist_valid = false;
				}
			}
			if (hist_valid) {
				if (hist_has_prev) {
					draw_vec(hist_x1,hist_prev_y,hist_x1,*yt);
				}
				draw_vec(hist_x1,*yt,hist_x2,*yt);
			}
			hist_has_prev = true;
			hist_prev_x = *xt; hist_prev_y = *yt;
		} else {
			hist_has_prev = false;
		}
	}
}

void do_draw_impulses(double* xt, double* yt, int* m, int npnts) {
	double impulses_orig = 0.0;
	if (wymin > 0.0) impulses_orig = wymin;
	if (wymax < 0.0) impulses_orig = wymax;
	for (int i = 0; i < npnts; i++, m++, xt++, yt++) {
		if (!*m) draw_vec(*xt, impulses_orig, *xt, *yt);
	}
}

void do_draw_bar(double* xt, double* yt, int* m, int npts) {
	do_draw_hist(xt, yt, m, npts);
	double hist_prev_x = 0, hist_prev_y = 0, hist_x1, hist_x2;
	bool hist_has_prev = false;
	double impulses_orig = 0.0;
	if (wymin > 0.0) impulses_orig = wymin;
	if (wymax < 0.0) impulses_orig = wymax;
	for (int i = 0; i < npts; i++, m++, xt++, yt++) {
		if (!*m) {
			bool hist_valid = true;
			if (i < npts-1 && !*(m+1)) {
				hist_x2 = (*(xt+1) + (*xt))/2;
				if (hist_has_prev) {
					hist_x1 = (hist_prev_x + (*xt))/2;
				} else {
					hist_x1 = 2*(*xt) - hist_x2;
				}
			} else {
				if (hist_has_prev) {
					hist_x1 = (hist_prev_x + (*xt))/2;
					hist_x2 = 2*(*xt) - hist_x1;
				} else {
					hist_valid = false;
				}
			}
			if (hist_valid && hist_has_prev) {
				// g_set_color(GLE_COLOR_BLACK);
				double yval = hist_prev_y;
				if (fabs(hist_prev_y-impulses_orig) > fabs(*yt-impulses_orig)) {
					yval = *yt;
				}
				draw_vec(hist_x1,impulses_orig,hist_x1,yval);
			}
			hist_has_prev = true;
			hist_prev_x = *xt; hist_prev_y = *yt;
		} else {
			hist_has_prev = false;
		}
	}

}

void do_svg_smooth(double* xold, int size) {
 /* for now this is a quadratic or cubic and 7 point smoothing only */
	int i;
	double* xnew;
	xnew = (double*) calloc(size,sizeof(double));

	for (i =0 ;i <= size ;i++ )
	{
		if (i == 0 || i == 1 || i == size-2 ||  i == size-1)
		{
			/* can't do it*/
			xnew[i]=xold[i];
		}
		else if (i == 2 || i == size-3)
		{
			/* do five point */
			xnew[i] = (-3*xold[i-2]+12*xold[i-1]+17*xold[i]+12*xold[i+1]-3*xold[i+2])/35;
		}
		else if (i == 3 || i == size - 4)
		{
			/* do seven point */
			xnew[i] = (-2*xold[i-3]+3*xold[i-2]+6*xold[i-1]+7*xold[i]+6*xold[i+1]+3*xold[i+2]-2*xold[i+3])/21;
		}
		else if (i >= 4 && i <= size - 5)
		{
			/* do nine point */
			xnew[i] = (-21*xold[i-4]+14*xold[i-3]+39*xold[i-2]+54*xold[i-1]+59*xold[i]+54*xold[i+1]+39*xold[i+2]+14*xold[i+3]-21*xold[i+4])/231;
		}

	}
	memcpy(xold,xnew,size*sizeof(double));
	free(xnew);
}

void big_vec(int nomiss) {
	double x1,y1,x2,y2;
	int m1,m2;


	if (fptr==NULL) return;
aaa:
	if (feof(fptr))  goto xxx;
	big_line(&x1,&y1,&m1);
	if (m1) goto aaa;
	for (;;) {
aaa2:
		if (feof(fptr))  break;
		big_line(&x2,&y2,&m2);
		if (nomiss && m2) goto aaa2;
		if (!m1 && !m2) draw_vec(x1,y1,x2,y2);
		x1 = x2; y1 = y2; m1 = m2;
	}
xxx:	fclose(fptr); fptr = NULL;
}

void big_close(void) {
	if (fptr!=NULL) fclose(fptr);
	fptr = NULL;
}

void big_mark(int mnum, double msize) throw (ParserError) {
	double x1,y1;
	int m2;

	if (fptr==NULL) return;
	if (feof(fptr))  goto xxx;
	for (;;) {
aaa2:
		if (feof(fptr))  break;
		big_line(&x1,&y1,&m2);
		if (m2) goto aaa2;
		draw_mark(x1,y1,mnum,msize,1.0);
	}
xxx:	fclose(fptr); fptr=NULL;
}

void g_move_safe(double x, double y) {
	if (gle_isnan(x)) return;
	if (gle_isnan(y)) return;
	g_move(x, y);
}

void g_line_safe(double x, double y) {
	if (gle_isnan(x)) return;
	if (gle_isnan(y)) return;
	g_line(x, y);
}

void draw_vec(double x1,double y1,double x2,double y2) {
	if (x1 < wxmin || x1 > wxmax || y1 < wymin || y1 > wymax ||
	    x2 < wxmin || x2 > wxmax || y2 < wymin || y2 > wymax) {
		/* ok one or both are outside our box */
		double txmin = wxmin, tymin = wymin;
		double txmax = wxmax, tymax = wymax;
		if (xx[2].log) {
			y1 = log10(y1); y2 = log10(y2);
			tymin = log10(tymin); tymax = log10(tymax);
		}
		if (xx[1].log) {
			x1 = log10(x1); x2 = log10(x2);
			txmin = log10(txmin); txmax = log10(txmax);
		}
		bool invis = gclip(&x1, &y1, &x2, &y2, txmin, tymin, txmax, tymax);
		if (invis) return;
		if (xx[2].log) {
			y1 = pow(10.0,y1); y2 = pow(10.0,y2);
		}
		if (xx[1].log) {
			x1 = pow(10.0,x1); x2 = pow(10.0,x2);
		}
	}
	if (x1!=last_vecx || y1!=last_vecy) {
		g_move_safe(fnx(x1), fny(y1));
	}
	g_line_safe(fnx(x2), fny(y2));
	last_vecx = x2;
	last_vecy = y2;
}

void draw_markers() throw (ParserError) {
	double *yd = NULL, dval, oldlwidth;
	char oldlstyle[10];
	g_gsave();
	g_get_line_style(oldlstyle);
	g_get_line_width(&oldlwidth);
	for (int dn = 1; dn <= ndata; dn++) {
		bool big = false;
		if (dp[dn] != NULL && dp[dn]->marker != 0) {
			if (dp[dn]->bigfile!=NULL) {
				big=true;
				big_open(dp[dn]->bigfile);
			} else {
				big = false;
			}
			g_set_line_width(oldlwidth); /* use defaults for each */
			g_set_color(dp[dn]->color);
			g_set_line_width(dp[dn]->lwidth);
			windowdn(dn);
			done_line = true;
			double* yt = dp[dn]->yv;
			double* xt = dp[dn]->xv;
			int* m = dp[dn]->miss;
			double msize = dp[dn]->msize;
			if (msize == 0) msize = g_fontsz;
			if (dp[dn]->mscale != 0) msize = msize * dp[dn]->mscale;
			if (big) {
				big_mark(dp[dn]->marker,msize);
			} else {
				double mdist = dp[dn]->mdist;
				if (mdist == 0) {
					if (dp[dn]->mdata!=0) yd = dp[dp[dn]->mdata]->yv;
					for (int i=0; i<dp[dn]->np; i++,m++,xt++,yt++,yd++) {
						if (!*m) {
							if (dp[dn]->mdata==0) dval = 1; else dval = *yd;
							draw_mark(*xt,*yt,dp[dn]->marker,msize,dval);
						}
					}
				} else {
					gr_nomiss(dn);
					if (dp[dn]->np >= 1) {
						double len = 0.0;
						double x0 = fnx(xt[0]);
						double y0 = fny(yt[0]);
						for (int i = 1; i < dp[dn]->np; i++) {
							double x = fnx(xt[i]);
							double y = fny(yt[i]);
							len += sqrt((x-x0)*(x-x0)+(y-y0)*(y-y0));
							x0 = x; y0 = y;
						}
						x0 = fnx(xt[0]);
						y0 = fny(yt[0]);
						double prev_dist = mdist - fmod(len, mdist)/2.0;
						for (int i = 1; i < dp[dn]->np; i++) {
							double x = fnx(xt[i]);
							double y = fny(yt[i]);
							double dist = sqrt((x-x0)*(x-x0)+(y-y0)*(y-y0));
							while (prev_dist + dist > mdist) {
								double offs = mdist - prev_dist;
								double xp = (offs*x + (dist-offs)*x0)/dist;
								double yp = (offs*y + (dist-offs)*y0)/dist;
								g_move(xp, yp);
								g_marker2(dp[dn]->marker,msize,1);
								x0 = xp; y0 = yp;
								dist = sqrt((x-x0)*(x-x0)+(y-y0)*(y-y0));
								prev_dist = 0.0;
							}
							prev_dist += dist;
							x0 = x; y0 = y;
						}
					}
				}
			}
			windownorm();
		}
	}
	g_grestore();
}

void draw_mark(double x, double y,int mrk, double msize, double dval) throw (ParserError) {
	/* globals   *** xmin, xmax, ymin, ymax,  ox,oy,  xbl, ybl */
	if (x >= wxmin && x<=wxmax && y>=wymin && y<=wymax) {
		g_move(fnx(x),fny(y));
		g_marker2(mrk,msize,dval);
	}
}

void box_clip(double *x,double *y,double wxmin,double wymin,double wxmax,double wymax) {
	if (*x > wxmax) *x = wxmax;
	if (*y > wymax) *y = wymax;
	if (*x < wxmin) *x = wxmin;
	if (*y < wymin) *y = wymin;
}

void gclip_simple(double *v, double vmin, double vmax) {
	if (*v < vmin) { *v = vmin; }
	if (*v > vmax) { *v = vmax; }
}

bool gclip(double *x1, double *y1, double *x2, double *y2,
	   double xmin, double ymin, double xmax, double ymax) {
	// Take care of infinite values
	if (gle_isinf(*y1)) {
		gclip_simple(y1, ymin, ymax);
		if (gle_isinf(*y2)) {
			if (*y1 == *y2) return true;
			else gclip_simple(y2, ymin, ymax);
		} else {
			*x1 = *x2;
		}
	} else if (gle_isinf(*y2)) {
		gclip_simple(y2, ymin, ymax);
		*x2 = *x1;
	}
	if (gle_isinf(*x1)) {
		gclip_simple(x1, xmin, xmax);
		if (gle_isinf(*x2)) {
			if (*x1 == *x2) return true;
			else gclip_simple(x2, xmin, xmax);
		} else {
			*y1 = *y2;
		}
	} else if (gle_isinf(*x2)) {
		gclip_simple(x2, xmin, xmax);
		*y2 = *y1;
	}
	// Check upper bound for all variables
	// Check X
	if (*x2 > xmax) {
		if (*x1 > xmax) return true;
		double dx = *x2 - *x1;
		if (dx == 0) return true;
		double dy = *y2 - *y1;
		*y2 = *y1 + dy*(xmax-*x1)/dx;
		*x2 = xmax;
	}
	if (*x1 > xmax) {
		double dx = *x1 - *x2;
		if (dx == 0) return true;
		double dy = *y1 - *y2;
		*y1 = *y2 + dy*(xmax-*x2)/dx;
		*x1 = xmax;
	}
	// Check Y
	if (*y2 > ymax) {
		if (*y1 > ymax) return true;
		double dy = *y2 - *y1;
		if (dy == 0) return true;
		double dx = *x2 - *x1;
		*x2 = *x1 + dx*(ymax-*y1)/dy;
		*y2 = ymax;
	}
	if (*y1 > ymax) {
		double dy = *y1 - *y2;
		if (dy == 0) return true;
		double dx = *x1 - *x2;
		*x1 = *x2 + dx*(ymax-*y2)/dy;
		*y1 = ymax ;
	}
	// Check lower bound for all variables
	// Check X
	if (*x2 < xmin) {
		if (*x1 < xmin) return true;
		double dx = *x2 - *x1;
		if (dx == 0) return true;
		double dy = *y2 - *y1;
		*y2 = *y1 + dy*(xmin-*x1)/dx;
		*x2 = xmin;
	}
	if (*x1 < xmin) {
		double dx = *x1 - *x2 ;
		if (dx == 0) return true;
		double dy = *y1 - *y2 ;
		*y1 = *y2 + dy*(xmin-*x2)/dx;
		*x1 = xmin;
	}
	// Check Y
	if (*y2 < ymin) {
		if (*y1 < ymin) return true;
		double dy = *y2 - *y1;
		if (dy == 0) return true;
		double dx = *x2 - *x1;
		*x2 = *x1 + dx*(ymin-*y1)/dy;
		*y2 = ymin;
	}
	if (*y1 < ymin) {
		double dy = *y1 - *y2;
		if (dy == 0) return true;
		double dx = *x1 - *x2;
		*x1 = *x2 + dx*(ymin-*y2)/dy;
		*y1 = ymin;
	}
	return false;
}

void vinit_axis(int i) {
	xx[i].init(i);
}

void vinit_title_axis() {
	xx[GLE_AXIS_T].log = 0;
	xx[GLE_AXIS_T].min = 0;
	xx[GLE_AXIS_T].max = 1;
	xx[GLE_AXIS_T].label_off = 1;
	xx[GLE_AXIS_T].side_off = 1;
	xx[GLE_AXIS_T].ticks_off = 1;
	xx[GLE_AXIS_T].subticks_off = 1;
	xx[GLE_AXIS_T].title_off = 0;
}

void doskip(char *s,int *ct) {
	if (*s==' ') (*ct)++;
}

double get_next_exp(TOKENS tk,int ntok,int *curtok) {
	static int elen,etype,cp,i;
	static double x;

	(*curtok)++;
	cp = 0;
	elen = 0;
	etype = 1;
	dbg for (i=1;i<=ntok;i++)  gprint("{%s} ",(*tk)[i]);
	dbg gprint("\n");
	dbg gprint("**get exp token ct %d  {%s} \n",*curtok,(*tk)[*curtok]);
	if (strlen(tk[*curtok])==0){
		/* this is added for optional parameters on smoothing. They must be at end for this to work*/
		dbg gprint("zero length expression in get expression no polish called\n");
		x=0;
	}else{
		polish_eval(tk[*curtok],&x);
	}
	return x;
}

void get_next_exp_string(TOKENS tk, int ntok, int *curtok, string* res) {
	(*curtok)++;
	if ((*curtok) <= ntok) {
		polish_eval_string(tk[*curtok],res);
	} else {
		*res = "";
	}
}

void get_next_exp_file(TOKENS tk, int ntok, int *curtok, string* res) {
	(*curtok)++;
	if ((*curtok) <= ntok) {
		pass_file_name(tk[*curtok], *res);
	} else {
		*res = "";
	}
}

bool checktok(char *t, char *want) {
	if (str_i_equals(t,want)) return true;
	else {
		gprint("Found token {%s} Wanted {%s} \n",t,want);
		return false;
	}
}

class DataFill {
protected:
	int m_Size;
	int m_VarX;
	char ostr[20];
	bool m_PrevNAN;
	bool m_PrevPoint;
	bool m_FineTune;
	double m_PrevXValue;
	double m_YMin;
	double m_YMax;
	GLEFunctionParserPcode* m_Expr;
	RefCountPtr<GLEDoubleArray> xarr;
	RefCountPtr<GLEDoubleArray> yarr;
	RefCountPtr<GLEBoolArray>   marr;
public:
	DataFill(bool finetune);
	void addPointIPol(double x);
	void addPoint(double xv, double yv);
	void addMissing(double xv);
	void tryIPol(double other, double nanval);
	void setInfo(GLEFunctionParserPcode* expr, int varx);
	void setYMinYMax(double ymin, double ymax);
	double computeYValue(double x);
	inline int size() { return m_Size; }
	inline double* getX() { return xarr->toArray(); }
	inline double* getY() { return yarr->toArray(); }
	inline int* getM() { return marr->toArray(); }
};

DataFill::DataFill(bool finetune) {
	m_Size = 0;
	m_PrevNAN = false;
	m_PrevPoint = false;
	m_FineTune = finetune;
	xarr = new GLEDoubleArray();
	yarr = new GLEDoubleArray();
	marr = new GLEBoolArray();
	m_YMin = graph_ymin - (graph_ymax - graph_ymin)/100.0;
	m_YMax = graph_ymax + (graph_ymax - graph_ymin)/100.0;
	if (finetune && graph_xmax == graph_xmin) {
		gprint("Graph scale not yet set before let command\n");
		return;
	}
}

void DataFill::setYMinYMax(double ymin, double ymax) {
	m_YMin = ymin - (ymax - ymin)/100.0;
	m_YMax = ymax + (ymax - ymin)/100.0;
}

void DataFill::setInfo(GLEFunctionParserPcode* expr, int varx) {
	m_Expr = expr; m_VarX = varx;
}

void DataFill::addPoint(double xv, double yv) {
	if (!gle_isnan(yv) && !gle_isinf(yv)) {
		setrange(xv, yv, 0);
		xarr->setDoubleAt(xv, m_Size);
		yarr->setDoubleAt(yv, m_Size);
		marr->setBoolAt(false, m_Size);
		m_Size++;
	}
}

void DataFill::addMissing(double xv) {
	xarr->setDoubleAt(xv, m_Size);
	yarr->setDoubleAt(0.0, m_Size);
	marr->setBoolAt(true, m_Size);
	m_Size++;
}

double DataFill::computeYValue(double x) {
	if (m_VarX >= 0) var_set(m_VarX, x);
	return m_Expr->evalDouble();
}

void DataFill::tryIPol(double other, double nanval) {
	int iter = 0;
	double limit = 0.001;
	while (true) {
		double xmid = (other + nanval)/2;
		double ymid = computeYValue(xmid);
		// cout << "xmid = " << xmid << " ymid = " << ymid << endl;
		if (gle_isnan(ymid) || ymid < m_YMin || ymid > m_YMax) {
			nanval = xmid;
		} else {
			other = xmid;
		}
		if (fabs(graph_xgraph(other) - graph_xgraph(nanval)) < limit) {
			addPoint(xmid, ymid);
			return;
		}
		iter++;
	}
}

void DataFill::addPointIPol(double x) {
	double y = computeYValue(x);
	// cout << "x = " << x << " y = " << y << endl;
	bool is_nan = gle_isnan(y);
	if (m_FineTune) {
		if (is_nan || y < m_YMin || y > m_YMax) {
			if (!m_PrevNAN && m_PrevPoint) {
				tryIPol(m_PrevXValue, x);
			}
			m_PrevNAN = true;
			addMissing(x);
			addPoint(x, y);
		} else {
			if (m_PrevNAN) {
				tryIPol(x, m_PrevXValue);
				m_PrevNAN = false;
			}
			addPoint(x, y);
		}
	} else {
		addPoint(x, y);
	}
	m_PrevPoint = true;
	m_PrevXValue = x;
}

class GLECheckWindow {
protected:
	bool m_HasXMin, m_HasXMax, m_HasYMin, m_HasYMax;
	double m_XMin, m_XMax, m_YMin, m_YMax;
public:
	GLECheckWindow();
	~GLECheckWindow();
	void setXMin(double v);
	void setXMax(double v);
	void setYMin(double v);
	void setYMax(double v);
	bool valid(double x, double y);
};

GLECheckWindow::GLECheckWindow() {
	m_HasXMin = m_HasXMax = m_HasYMin = m_HasYMax = false;
	m_XMin = m_XMax = m_YMin = m_YMax = 0.0;
}

GLECheckWindow::~GLECheckWindow() {
}

void GLECheckWindow::setXMin(double v) {
	m_XMin = v;
	m_HasXMin = true;
}

void GLECheckWindow::setXMax(double v) {
	m_XMax = v;
	m_HasXMax = true;
}

void GLECheckWindow::setYMin(double v) {
	m_YMin = v;
	m_HasYMin = true;
}

void GLECheckWindow::setYMax(double v) {
	m_YMax = v;
	m_HasYMax = true;
}

bool GLECheckWindow::valid(double x, double y) {
	if (m_HasXMin && x < m_XMin) return false;
	if (m_HasXMax && x > m_XMax) return false;
	if (m_HasYMin && y < m_YMin) return false;
	if (m_HasYMax && y < m_YMax) return false;
	return true;
}

class GLEFitLS : public GLEPowellFunc {
protected:
	int m_IdxX, m_NIter;
	double m_RSquare;
	vector<int> m_Vars;
	vector<double>* m_X;
	vector<double>* m_Y;
	GLEFunctionParserPcode m_Function;
	StringIntHash m_VarMap;
	string m_FunctionStr;
public:
	GLEFitLS();
	virtual ~GLEFitLS();
	void polish(const string& str) throw(ParserError);
	void setXY(vector<double>* x, vector<double>* y);
	void fit();
	void testFit();
	void setVarsVals(double* vals);
	void toFunctionStr(const string& format, string* str) throw(ParserError);
	virtual double fitMSE(double* vals);
	inline GLEFunctionParserPcode* getFunction() { return &m_Function; }
	inline double getRSquare() { return m_RSquare; }
};

GLEFitLS::GLEFitLS() {
	m_IdxX = -1;
	m_NIter = 0;
	m_RSquare = 0.0;
}

GLEFitLS::~GLEFitLS() {
}

void GLEFitLS::polish(const string& str) throw(ParserError) {
	m_FunctionStr = str;
	m_Function.polish(str.c_str(), &m_VarMap);
	/* Iterate over variables in expression */
	for (StringIntHash::const_iterator i = m_VarMap.begin(); i != m_VarMap.end(); i++ ) {
		if (i->first != "X") {
			m_Vars.push_back(i->second);
		}
	}
}

void GLEFitLS::setXY(vector<double>* x, vector<double>* y) {
	m_X = x;
	m_Y = y;
}

void GLEFitLS::fit() {
	int naz = m_Vars.size();
	double** xi = matrix(1,naz,1,naz);
	for (int i = 1; i <= naz; i++) {
		for (int j = 1; j <= naz; j++) {
			xi[i][j] = 0;
		}
		xi[i][i] = 1;
	}
	double* pms = new double[naz+1];
	for (int i = 1; i <= naz; i++) {
		int v_idx = m_Vars[i-1];
		var_get(v_idx, &pms[i]);
	}
	int vtype;
	double fret = 0;
	double anstol = 1e-4;
	var_findadd("X", &m_IdxX, &vtype);
	powell(pms, xi, naz, anstol, &m_NIter, &fret, this);
	free_matrix(xi,1,naz,1,naz);
	setVarsVals(pms);
}

void GLEFitLS::testFit() {
	int nxy = m_X->size();
	double sumy = 0.0;
	for (int i = 0; i < nxy; i++) {
		sumy = sumy + (*m_Y)[i];
	}
	double meany = sumy/nxy;
	double sum1 = 0.0, sum2 = 0.0;
	for (int i = 0; i < nxy; i++) {
		var_set(m_IdxX, (*m_X)[i]);
		double value = m_Function.evalDouble();
		double y = (*m_Y)[i];
		double r1 = value - y;
		double r2 = meany - y;
		sum1 = sum1 + r1*r1;
		sum2 = sum2 + r2*r2;
	}
	m_RSquare = 1 - sum1/sum2;
}

void GLEFitLS::setVarsVals(double* vals) {
	/* Set all variables to given values */
	int naz = m_Vars.size();
	for (int i = 1; i <= naz; i++) {
		int v_idx = m_Vars[i-1];
		if (v_idx >= 0) var_set(v_idx, vals[i]);
	}
}

double GLEFitLS::fitMSE(double* vals) {
	/* Set all variables to given values */
	setVarsVals(vals);
	/* Compute MSE */
	double tot = 0.0;
	for (vector<double>::size_type i = 0; i < m_X->size(); i++) {
		var_set(m_IdxX, (*m_X)[i]);
		double value = m_Function.evalDouble();
		double residue = (*m_Y)[i] - value;
		tot += residue*residue;
	}
	/* Return error */
	return tot / m_X->size();
}

void GLEFitLS::toFunctionStr(const string& format, string* str) throw(ParserError) {
	*str = "";
	string fmt_str = format;
	if (fmt_str == "") fmt_str = "fix 3";
	GLENumberFormat fmt(fmt_str);
	GLEPolish* polish = get_global_polish();
	Tokenizer* tokens = polish->getTokens(m_FunctionStr);
	string uc_token, v_str;
	bool has_plus = false;
	while (tokens->has_more_tokens()) {
		const string& token = tokens->next_token();
		str_to_uppercase(token, uc_token);
		int v_idx = m_VarMap.try_get(uc_token);
		if (uc_token != "X" && v_idx != -1) {
			double value;
			var_get(v_idx, &value);
			fmt.format(value, &v_str);
			if (has_plus && value >= 0) *str = *str + "+";
			*str = *str + v_str;
			has_plus = false;
		} else {
			if (has_plus) *str = *str + "+";
			has_plus = token == "+";
			if (!has_plus) *str = *str + token;
		}
	}
}

class GLELet {
protected:
	GLEVarSubMap* m_SubMap;
	double m_LetFrom, m_LetTo, m_LetStep, m_LogStep;
	bool m_FineTune, m_HasSteps, m_HasFrom, m_HasTo;
	int m_VarX;
public:
	GLELet();
	~GLELet();
	void initVars();
	void initStep();
	void doLet(GLEFunctionParserPcode* fct, int dset_id) throw(ParserError);
	void doFitFunction(int fitds, const string& fct, GLEParser* parser, bool finetune) throw(ParserError);
	void doHistogram(int histds, GLEParser* parser) throw(ParserError);
	void setStep(double value);
	inline void setVarSubMap(GLEVarSubMap* map) { m_SubMap = map; }
	inline void setFrom(double value) { m_LetFrom = value; }
	inline void setTo(double value) { m_LetTo = value; }
	inline void setFineTune(bool value) { m_FineTune = value; }
	inline double getFrom() { return m_LetFrom; }
	inline double getTo() { return m_LetTo; }
	inline double getStep() { return m_LetStep; }
	inline bool hasSteps() { return m_HasSteps; }
	inline bool hasFrom() { return m_HasFrom; }
	inline bool hasTo() { return m_HasTo; }
	inline void setHasSteps(bool has) { m_HasSteps = has; }
	inline void setHasFrom(bool has) { m_HasFrom = has; }
	inline void setHasTo(bool has) { m_HasTo = has; }
};

GLELet::GLELet() {
	m_SubMap = NULL;
	m_LetFrom = 0.0;
	m_LetTo = 0.0;
	m_LetStep = 0.0;
	m_VarX = -1;
	m_FineTune = false;
	m_HasSteps = false;
	m_HasFrom = false;
	m_HasTo = false;
}

GLELet::~GLELet() {
}

void GLELet::setStep(double value) {
	m_LetStep = value;
	setHasSteps(true);
}

void GLELet::initVars() {
	int vartype = 1;
	var_findadd("X", &m_VarX, &vartype);
}

void GLELet::initStep() {
	if (hasSteps()) return;
	if (xx[1].log) {
		setStep(DEFAULT_STEPS);
	} else {
		setStep((getTo() - getFrom())/DEFAULT_STEPS);
	}
}

void GLELet::doLet(GLEFunctionParserPcode* fct, int dset_id) throw(ParserError) {
	double logstep = 1.0;
	int ndn = 0;
	int dn_idx[11], dn_var[11];
	if (m_SubMap != NULL) {
		var_find_dn(m_SubMap, dn_idx, dn_var, &ndn);
	}
	if (m_LetTo <= m_LetFrom) {
		stringstream ss;
		ss << "range of let expression is empty: from = " << m_LetFrom << " to = " << m_LetTo;
		g_throw_parser_error(ss.str());
	}
	if (ndn == 0 && xx[1].log) {
		if (m_LetStep < 2) {
			stringstream ss;
			ss << "with a LOG xaxis scale STEP is taken as the number of steps n," << endl;
			ss << "which should be at least 2, but found: " << m_LetStep;
			g_throw_parser_error(ss.str());
		}
		logstep = pow(m_LetTo/m_LetFrom, 1.0/(m_LetStep-1));
	}
	if (ndata<dset_id) ndata = dset_id;
	if (dp[dset_id]==NULL) {
		dp[dset_id] = DP_CAST myallocz(sizeof(*dp[dset_id]));
		copy_default(dset_id);
	}
	/* copy default dataset settings */
	if (dp[dset_id]==NULL) gprint("Memory allocation error, graph dataset \n");
	DataFill fill(m_FineTune);
	if (dp[dset_id]->ymax > dp[dset_id]->ymin) {
		/* setting range based on range for specific data set */
		fill.setYMinYMax(dp[dset_id]->ymin, dp[dset_id]->ymax);
	}
	if (ndn > 0) {
		int dn = dn_var[0];
		double* xf = dp[dn]->xv;
		int* mm = dp[dn]->miss;
		for (int i = 0; i < dp[dn]->np; i++, xf++, mm++) {
			if ((!m_HasFrom || *xf >= m_LetFrom) && (!m_HasTo || *xf <= m_LetTo)) {
				if (m_VarX >= 0) var_set(m_VarX, *xf);
				for (int j = 0; j < ndn; j++) {
					if (dp[dn_var[j]] == NULL) {
						gprint("Dataset not defined {%d} \n",dn_var[j]);
					} else {
						var_set(dn_idx[j],dp[dn_var[j]]->yv[i]);
					}
				}
				if (! *mm) {
					double y = fct->evalDouble();
					fill.addPoint(*xf, y);
				}
			}
	 	}
	} else {
		//
		// fill up data from function supplied byuser
		//
		fill.setInfo(fct, m_VarX);
		for (double ix = m_LetFrom; ix < m_LetTo; ) {
			fill.addPointIPol(ix);
			if (xx[1].log == static_cast<int>(true)) {
				ix *= logstep;
			} else {
				ix += m_LetStep;
			}
		}
		fill.addPointIPol(m_LetTo);
	}
	dp[dset_id]->np = fill.size();
	if (dp[dset_id]->np == 0) {
		g_throw_parser_error("no data points in data set d", dset_id);
	}
	dp[dset_id]->miss = fill.getM();
	dp[dset_id]->xv = fill.getX();
	dp[dset_id]->yv = fill.getY();
}

void GLELet::doFitFunction(int fitds, const string& fct, GLEParser* parser, bool finetune) throw(ParserError) {
	string SlopeVar, OffsetVar, RSquaredVar;
	bool linfit = false, logefit = false, log10fit = false, powxfit = false, genfit = false, SetVars = false;
	Tokenizer* tokens = parser->getTokens();

	// doing fitting routines
	if (str_i_equals(fct,"LINFIT"))   linfit = true;
	if (str_i_equals(fct,"LOGEFIT"))  logefit = true;
	if (str_i_equals(fct,"LOG10FIT")) log10fit = true;
	if (str_i_equals(fct,"POWXFIT"))  powxfit = true;
	if (str_i_equals(fct,"FIT"))      genfit = true;

	// compose a new let string and recurse to this function
	// get the data series to linear fit to
	string& token = tokens->next_token();
	if (token.length() <= 1 || toupper(token[0]) != 'D') {
		throw tokens->error("data set identifier expected after fit command");
	}
	int ddlinfit = get_dataset_identifier(token.c_str(), false);

	// do the fit
	// check the next tokens
	vector<double> x, y;
	double xmax=0.0,xmin=0.0,ymax=0.0,ymin=0.0;
	bool limit_data_y=false;  // plot as far as y data
	bool limit_data_x=false;  // plot as far as x data
	bool limit_data=false; // plot which ever is maximum x or y
	string fitfct, eqstr, format, rsq;

	GLECheckWindow window;
	while (true) {
		 string& token = tokens->try_next_token();
		if (str_i_equals(token, "WITH")) {
			fitfct = tokens->next_multilevel_token();
		} else if (str_i_equals(token, "EQSTR")) {
			eqstr = tokens->next_token();
		} else if (str_i_equals(token, "FORMAT")) {
			format = tokens->next_token();
			str_remove_quote(format);
		} else if (str_i_equals(token, "RSQ")) {
			rsq = tokens->next_token();
		} else if (str_i_equals(token, "FROM")) {
			setHasFrom(true);
			setFrom(parser->evalTokenToDouble());
		} else if (str_i_equals(token,"TO")) {
			setHasTo(true);
			setTo(parser->evalTokenToDouble());
		} else if (str_i_equals(token, "STEP")) {
			setStep(parser->evalTokenToDouble());
		} else if (str_i_equals(token, "LIMIT_DATA_X")) {
			// user wants the data ploted from dd xmax to dd xmin
			limit_data_x = true;
		} else if (str_i_equals(token, "LIMIT_DATA_Y")) {
			// get x values from evaluation of slope
			limit_data_y = true;
		} else if (str_i_equals(token, "LIMIT_DATA")) {
			limit_data = true;
		} else if (str_i_equals(token, "XMIN")) {
			double xm = parser->evalTokenToDouble();
			window.setXMin(xm);
			setFrom(xm);
		} else if (str_i_equals(token, "XMAX")) {
			double xm = parser->evalTokenToDouble();
			window.setXMax(xm);
			setTo(xm);
		} else if (str_i_equals(token, "YMIN")) {
			window.setYMin(parser->evalTokenToDouble());
		} else if (str_i_equals(token, "YMAX")) {
			window.setYMax(parser->evalTokenToDouble());
		} else {
			if (token != "") tokens->pushback_token();
			break;
		}
	}

	// copy to vectors but ignore miss
	bool notset = true;
	for (int i=0;i<dp[ddlinfit]->np;i++) {
		if (!dp[ddlinfit]->miss[i]) {
			if (window.valid(dp[ddlinfit]->xv[i], dp[ddlinfit]->yv[i])) {
				if (notset) {
					xmin = xmax = dp[ddlinfit]->xv[i];
					ymin = ymax = dp[ddlinfit]->yv[i];
					notset = false;
				}
				xmax = max(xmax,dp[ddlinfit]->xv[i]);
				xmin = min(xmin,dp[ddlinfit]->xv[i]);
				ymax = max(ymax,dp[ddlinfit]->yv[i]);
				ymin = min(ymin,dp[ddlinfit]->yv[i]);
				x.push_back(dp[ddlinfit]->xv[i]);
				y.push_back(dp[ddlinfit]->yv[i]);
			}
		}
	}

	if(limit_data_x){
		setFrom(xmin);
		setTo(xmax);
	}

	// user can supply three variables to get the fit: SLOPE OFFSET R_SQUARED
	if (tokens->has_more_tokens()) {
		SlopeVar = tokens->next_token();
		ensure_valid_var_name(tokens, SlopeVar);
		SetVars = true;
	}
	if (tokens->has_more_tokens()) {
		OffsetVar = tokens->next_token();
		ensure_valid_var_name(tokens, OffsetVar);
		SetVars = true;
	}
	if (tokens->has_more_tokens()) {
		RSquaredVar = tokens->next_token();
		ensure_valid_var_name(tokens, RSquaredVar);
		SetVars = true;
	}

	if (tokens->has_more_tokens()) {
		throw tokens->error("extra tokens at end of let command");
	}

	double slope=0,offset=0,rsquared=0,temp_xmax=0,temp_xmin=0;
	char* new_let = new char[1000];
	if (linfit) {
		least_square(&x,&y,&slope,&offset,&rsquared);
		if (limit_data_y || limit_data) {
			// find x values from ymax and min
			// y=slope*x+offset
			// x=(y-offset)/slope
			temp_xmax = max( (ymax-offset)/slope , (ymin-offset)/slope);
			temp_xmin = min( (ymax-offset)/slope , (ymin-offset)/slope);
			if (limit_data_y) {
				setTo(temp_xmax);
				setFrom(temp_xmin);
			} else if (limit_data) {
				setTo(max(temp_xmax,xmax));
				setFrom(min(temp_xmin,xmin));
			}
		}
		sprintf(new_let,"let d%d = %0.10e*x+%0.10e FROM %0.10e TO %0.10e",fitds,slope,offset,getFrom(),getTo());
	} else if (logefit) {
		//take loge of y values
		vector<double>::iterator vdi=y.begin();
		while (vdi != y.end()) {
			(*vdi) = log(*vdi);
			vdi++;
		}
		least_square(&x,&y,&slope,&offset,&rsquared);
		if (limit_data_y || limit_data) {
			// find x values from ymax and min
			// y=exp(offset)*exp(slope*x)
			// x=(log(y)-offset)/slope
			//
			temp_xmax = max( (log(ymax)-offset)/slope , (log(ymin)-offset)/slope);
			temp_xmin = min( (log(ymax)-offset)/slope , (log(ymin)-offset)/slope);
			if (limit_data_y) {
				setTo(temp_xmax);
				setFrom(temp_xmin);
			} else if (limit_data) {
				setTo(max(temp_xmax,xmax));
				setFrom(min(temp_xmin,xmin));
			}
		}
		offset = exp(offset);
		sprintf(new_let,"let d%d = %0.10e*exp(%0.10e*x) FROM %0.10e TO %0.10e",fitds,offset,slope,getFrom(),getTo());
	} else if (log10fit) {
		//take log10 of y values
		vector<double>::iterator vdi=y.begin();
		while (vdi != y.end()) {
			(*vdi) = log10(*vdi);
			vdi++;
		}
		least_square(&x,&y,&slope,&offset,&rsquared);
		if (limit_data_y || limit_data) {
			// find x values from ymax and min
			// y=exp(offset)*exp(slope*x)
			// x=(log10(y)-offset)/slope
			//
			temp_xmax = max( (log10(ymax)-offset)/slope , (log10(ymin)-offset)/slope);
			temp_xmin = min( (log10(ymax)-offset)/slope , (log10(ymin)-offset)/slope);
			if (limit_data_y) {
				setTo(temp_xmax);
				setFrom(temp_xmin);
			} else if (limit_data) {
				setTo(max(temp_xmax,xmax));
				setFrom(min(temp_xmin,xmin));
			}
		}
		offset = pow(10.0,offset);
		sprintf(new_let,"let d%d = %0.10e*10^(%0.10e*x) FROM %0.10e TO %0.10e",fitds,offset,slope,getFrom(),getTo());
	} else if (powxfit) {
		//take loge of both axis
		vector<double>::iterator vdi=y.begin();
		while (vdi != y.end()) {
			(*vdi) = log(*vdi);
			vdi++;
		}
		vdi=x.begin();
		while (vdi != x.end()) {
			(*vdi) = log(*vdi);
			vdi++;
		}
		least_square(&x,&y,&slope,&offset,&rsquared);
		if (limit_data_y || limit_data) {
			// find x values from ymax and min
			// y=exp(offset)*x^(slope)
			// x = ( y / exp(offset) ) ^ (1/slope)
			temp_xmax = max( pow(ymax/exp(offset),1.0/slope) ,pow(ymin/exp(offset),1.0/slope) );
			temp_xmin = min( pow(ymax/exp(offset),1.0/slope) ,pow(ymin/exp(offset),1.0/slope) );
			if (limit_data_y) {
				setTo(temp_xmax);
				setFrom(temp_xmin);
			} else if (limit_data) {
				setTo(max(temp_xmax,xmax));
				setFrom(min(temp_xmin,xmin));
			}
		}
		offset = exp(offset);
		sprintf(new_let,"let d%d = %0.10e*x^(%0.10e) FROM %0.10e TO %0.10e",fitds,offset,slope,getFrom(),getTo());
	} else if (genfit) {
		GLEFitLS fitls;
		fitls.polish(fitfct);
		fitls.setXY(&x,&y);
		fitls.fit();
		fitls.testFit();
		initVars();
		initStep();
		doLet(fitls.getFunction(), fitds);
		if (eqstr != "") {
			string eqstr_v;
			str_to_uppercase(eqstr);
			fitls.toFunctionStr(format, &eqstr_v);
			var_findadd_set(eqstr.c_str(), eqstr_v);
		}
		if (rsq != "") {
			str_to_uppercase(rsq);
			var_findadd_set((char*)rsq.c_str(), fitls.getRSquare());
		}
		delete []new_let;
		return;
	}
	// add the variables -- they will be globals?
	if (SetVars){
		str_to_uppercase(SlopeVar);
		var_findadd_set((char*)SlopeVar.c_str(), slope);
		str_to_uppercase(OffsetVar);
		var_findadd_set((char*)OffsetVar.c_str(), offset);
		str_to_uppercase(RSquaredVar);
		var_findadd_set((char*)RSquaredVar.c_str(), rsquared);
	}
	// cout << "new let: " << new_let << endl;
	do_let(new_let, finetune);
	delete []new_let;
	return;
}

void GLELet::doHistogram(int resds, GLEParser* parser) throw(ParserError) {
	Tokenizer* tokens = parser->getTokens();
	// get the data series to compute histogram of
	string& token = tokens->next_token();
	if (token.length() <= 1 || toupper(token[0]) != 'D') {
		throw tokens->error("data set identifier expected after histogram command");
	}
	int bins = -1;
	int histds = get_dataset_identifier(token.c_str(), true);
	while (tokens->has_more_tokens()) {
		string& token = tokens->next_token();
		if (str_i_equals(token, "FROM")) {
			setHasFrom(true);
			setFrom(parser->evalTokenToDouble());
		} else if (str_i_equals(token,"TO")) {
			setHasTo(true);
			setTo(parser->evalTokenToDouble());
		} else if (str_i_equals(token, "STEP")) {
			setStep(parser->evalTokenToDouble());
		} else if (str_i_equals(token, "BINS")) {
			bins = (int)floor(parser->evalTokenToDouble()+0.5);
		} else {
			stringstream errstr;
			errstr << "unknown token in 'let' expression: '" << token << "'";
			throw tokens->error(errstr.str());
		}
	}
	axis_struct* ax = &xx[GLE_AXIS_X];
	if (!hasFrom() && ax->minset) {
		setHasFrom(true);
		setFrom(ax->min);
	}
	if (!hasTo() && ax->maxset) {
		setHasTo(true);
		setTo(ax->max);
	}
	if (!hasFrom() || !hasTo()) {
		range_y1 = 1e30; range_y2 = -1e30;
		int np = dp[histds]->np;
		double* yv = dp[histds]->yv;
		int* m = dp[histds]->miss;
		for (int i = 0; i < np; i++) {
			setyrange(yv[i], m[i]);
		}
		roundrange(&range_y1, &range_y2, false, false);
		if (range_y2 > range_y1) {
			setFrom(range_y1);
			setTo(range_y2);
		}
		range_y1 = 1e30; range_y2 = -1e30;
	}
	if (bins == -1 && !hasSteps()) {
		bins = 10;
	}
	vector<double> from;
	vector<double> counts;
	if (bins != -1) {
		for (int i = 0; i < bins; i++) {
			from.push_back(getFrom() + (getTo()-getFrom())*i/bins);
			counts.push_back(0.0);
		}
		from.push_back(getTo());
	} else {
		int i = 0;
		double value = getFrom();
		while (value < getTo()) {
			from.push_back(value);
			counts.push_back(0.0);
			i++;
			// avoids rounding errors
			value = getFrom()+i*getStep();
		}
		from.push_back(value);
	}
	for (int i=0; i < dp[histds]->np; i++) {
		if (!dp[histds]->miss[i]) {
			int found = -1;
			double yv = dp[histds]->yv[i];
			for (vector<double>::size_type j = 0; j < counts.size(); j++) {
				if (yv >= from[j] && yv < from[j+1]) {
					found = j;
					break;
				}
			}
			if (found != -1) counts[found]++;
			else if (yv == getTo()) counts[counts.size()-1]++;
		}
	}
	DataFill fill(false);
	for (vector<double>::size_type i = 0; i < from.size()-1; i++) {
		fill.addPoint((from[i]+from[i+1])/2.0, counts[i]);
	}
	dp[resds]->np = fill.size();
	dp[resds]->miss = fill.getM();
	dp[resds]->xv = fill.getX();
	dp[resds]->yv = fill.getY();
}

/*  LET d2 = exp(x) [ FROM exp TO exp [ STEP exp ] ]	*/
/*  LET d3 = exp(dn,d...)				*/
// added by V.L.
//	LET d3 = LINFIT(dn) [FROM exp TO exp] or [DATA] [SLOPE] [OFFSET] [R_SQUARED]
// [SLOPE] [OFFSET] [R_SQUARED] are user defined variables
//  no option after linfit results in line being drawn over the whole graph
// DATA results in the line being drawn from the minimum to the maximum of the data series
//
void do_let(int line, bool finetune) throw(ParserError) {
	string letcmd;
	g_set_error_line(line);
	get_block_line(line, letcmd);
	do_let(letcmd, finetune);
}

void do_let(const string& letcmd, bool finetune) throw(ParserError) {
	GLELet let;

	GLEParser* parser = get_global_parser();
	parser->setString(letcmd.c_str());
	Tokenizer* tokens = parser->getTokens();
	tokens->ensure_next_token_i("LET");

	let.setFineTune(finetune);

	// dd is the data set to assigned the let to
	// atoi the +1 strips the d
	string& token = tokens->next_token();
	if (token.length() <= 1 || toupper(token[0]) != 'D') {
		throw tokens->error("data set identifier expected after let command");
	}
	int dd = get_dataset_identifier(token.c_str());

	// next token should be an equal sign
	tokens->ensure_next_token("=");

	// default is to plot over the entire data range
	// cout << "from = " << xx[1].min << " to = " << xx[1].max << endl;
	let.setFrom(xx[1].min);
	let.setTo(xx[1].max);

	string let_fct = tokens->next_multilevel_token();
	if (str_i_equals(let_fct,"LINFIT")   ||
	    str_i_equals(let_fct,"LOGEFIT")  ||
	    str_i_equals(let_fct,"LOG10FIT") ||
	    str_i_equals(let_fct,"POWXFIT")  ||
	    str_i_equals(let_fct,"FIT")) {
		let.doFitFunction(dd, let_fct, parser, finetune);
		return;
	}

	if (str_i_equals(let_fct,"HIST")) {
		let.doHistogram(dd, parser);
		return;
	}

	// iterate over all tokens
	while (tokens->has_more_tokens()) {
		token = tokens->next_token();
		if (str_i_equals(token,"FROM")) {
			let.setHasFrom(true);
			let.setFrom(parser->evalTokenToDouble());
		} else if (str_i_equals(token,"TO")) {
			let.setHasTo(true);
			let.setTo(parser->evalTokenToDouble());
		} else if (str_i_equals(token,"STEP")) {
			let.setStep(parser->evalTokenToDouble());
		} else if (str_i_equals(token,"NOTUNE")) {
			let.setFineTune(false);
		} else {
			stringstream errstr;
			errstr << "unknown token in 'let' expression: '" << token << "'";
			throw tokens->error(errstr.str());
		}
	}

	let.initStep();
	GLEVarSubMap* submap = var_add_local_submap();
	GLEFunctionParserPcode function;
	let.initVars();
	function.polish(let_fct.c_str());
	let.setVarSubMap(submap);
	let.doLet(&function, dd);
	var_remove_local_submap();
}

extern key_struct *kd[100];

void gdraw_key(KeyInfo* info) {
	draw_key(g_nkd, info);
}

void next_lstyle(char* s,int* ct) {
	char next[200];
	double temp;
	int is_alpha;
	int i,len;

	is_alpha=false;
	(*ct)++;
	doskip(tk[*ct],ct);
	strcpy(next,tk[*ct]);
	len = strlen(next);

	for(i=0 ; i < len ; i++) {
		is_alpha=isalpha((int)next[i]);
		if (is_alpha) i = len;
	}
	if (is_alpha) {
		/* this is a variable so evaluate it */
		polish_eval(next,&temp);
		sprintf(s,"%g",temp);
	} else {
		/*		it is a number so just copy over*/
		if (len < 9) {
			strcpy(s,next);
		} else {
			gprint("ERROR line style string too long %s\n",next);
		}
	}
}

void next_svg_iter(int* s, int* ct) {
	char next[200];
	double temp;
	int is_alpha;
	int i,len,idx,type;

	is_alpha=false;
	(*ct)++;
	doskip(tk[*ct],ct);
	strcpy(next,tk[*ct]);
	len = strlen(next);
	printf("len=%d next=%s\n",len,next);
/*
	printf("\nnext=%s\n",next);
*/
	// check to see if number or alphanumeric
	if ( len > 0 )
	{
	for(i=0 ; i < len ; i++)
	{
		is_alpha=isalpha((int)next[i]);
		if (is_alpha) i = len;
	}
	if (is_alpha)
	{

		// it is alphanumeric so see if it is a variable
		// if not set it to zero
		var_find(next,&idx,&type);
		if (idx != -1 )
		{
			/* this is a variable so evaluate it */
			polish_eval(next,&temp);
			*s = (int) temp;
		}
		else
		{
			// its not a variable must be another keyword
			// set iter to 1
			(*ct)--;
			*s = 1;
		}
		//sprintf(s,"%g\0",temp);
	}
	else
	{
		/* it is a number so just copy over*/
		*s = atoi(next);
	}
	}
	else
	{
		//len == 0 so there isnothing there set to 1
		*s = 1;
		(*ct)--;
	}
}

#define next_dn  (ct+=1,skipspace,atoi(tk[ct]+1))

void do_dataset(int d) {
	int ct=2;
	//cout <<endl<<"d="<< d <<" ";
	dp[d]->axisscale = true;
	while (ct<=ntk)
	{
		//cout << "ct="<<ct<<"ntk="<<ntk<<" ";
	     kw("LINE") 	dp[d]->line = true;
	else kw("LSTYLE") 	next_lstyle(dp[d]->lstyle,&ct);
/* allow variable for line style */
/*	else kw("LSTYLE") 	dp[d]->lstyle = next_exp;*/
	else kw("LWIDTH") 	dp[d]->lwidth = next_exp;
	else kw("MARKER") 	dp[d]->marker = next_marker;
	else kw("MDATA") 	dp[d]->mdata = next_dn;
	else kw("COLOR") 	dp[d]->color = next_color;
	else kw("KEYFILL") 	dp[d]->key_fill = next_color;
	else kw("MSIZE") 	dp[d]->msize = next_exp;
	else kw("MDIST") 	dp[d]->mdist = next_exp;
	else kw("MSCALE") 	dp[d]->mscale = next_exp;
	else kw("KEY") 		next_vquote(dp[d]->key_name);
	else kw("AUTOSCALE") 	dp[d]->autoscale = true;
	else kw("AUTO") 	dp[d]->autoscale = true;
	else kw("NOMISS") 	dp[d]->nomiss = true;
	else kw("NOMISSING") 	dp[d]->nomiss = true;
	else kw("BIGFILE") 	next_vquote(dp[d]->bigfile);
	else kw("STEPS")        dp[d]->line_mode = GLE_GRAPH_LM_STEPS;
	else kw("FSTEPS")       dp[d]->line_mode = GLE_GRAPH_LM_FSTEPS;
	else kw("HIST")         dp[d]->line_mode = GLE_GRAPH_LM_HIST;
	else kw("BAR")          dp[d]->line_mode = GLE_GRAPH_LM_BAR;
	else kw("IMPULSES")     dp[d]->line_mode = GLE_GRAPH_LM_IMPULSES;
	else kw("X2AXIS")       dp[d]->axis = GLE_AXIS_X2;
	else kw("Y2AXIS")       dp[d]->axis = GLE_AXIS_Y2;
	else kw("SMOOTH") 	{
		//cout << "smooth";
		dp[d]->smoothm = false;
		dp[d]->smooth = true;
		dp[d]->line = true;
	}
	else kw("SMOOTHM") 	{
		dp[d]->smoothm = true;
		dp[d]->smooth = true;
		dp[d]->line = true;
	}
	else kw("SVG_SMOOTH") {
		//next_svg_iter(&dp[d]->svg_iter,&ct);
		dp[d]->svg_iter = (int) next_exp;
		if (dp[d]->svg_iter == 0) dp[d]->svg_iter = 1; // if they say svg_smooth they want it at least once
		//printf("iter = %d\n",dp[d]->svg_iter);
		dp[d]->svg_smooth = true;
		dp[d]->smoothm = false;
		dp[d]->smooth = false;
		dp[d]->line = true;
	}
	else kw("DERESOLVE") {
		dp[d]->deresolve = (int) next_exp;
		dp[d]->deresolve_avg = false;
		if (str_i_equals(tk[ct+1], "AVERAGE")) {
			dp[d]->deresolve_avg = true;
			ct++;
		}
	}
	else kw("XMIN") 	dp[d]->xmin = next_exp;
	else kw("XMAX") 	dp[d]->xmax = next_exp;
	else kw("YMIN") 	dp[d]->ymin = next_exp;
	else kw("YMAX") 	dp[d]->ymax = next_exp;
	else kw("HERR") {
		next_str(dp[d]->herrup);
		strcpy(dp[d]->herrdown, dp[d]->herrup);
	}
	else kw("HERRLEFT")	next_str(dp[d]->herrup);
	else kw("HERRRIGHT")	next_str(dp[d]->herrdown);
	else kw("HERRWIDTH")	dp[d]->herrwidth = next_exp;
	else kw("ERR") {
		next_str(dp[d]->errup);
		strcpy(dp[d]->errdown, dp[d]->errup);
	}
	else kw("ERRUP")	next_str(dp[d]->errup);
	else kw("ERRDOWN")	next_str(dp[d]->errdown);
	else kw("ERRWIDTH")	dp[d]->errwidth = next_exp;
	else gprint("Unrecognised GRAPH DN sub command {%s} \n ",tk[ct]);
	ct++;
	}
	do_dataset_key(d);
}

void do_dataset_key(int d) {
	if (dp[d] != NULL && dp[d]->key_name != NULL) {
		kd[++g_nkd] = new key_struct(g_keycol);
		kd[g_nkd]->fill = dp[d]->key_fill;
		kd[g_nkd]->color = dp[d]->color;
		kd[g_nkd]->lwidth = dp[d]->lwidth;
		kd[g_nkd]->marker = dp[d]->marker;
		kd[g_nkd]->msize = dp[d]->msize;
		strcpy(kd[g_nkd]->lstyle,dp[d]->lstyle);
		if (kd[g_nkd]->lstyle[0] == 0 && dp[d]->line) {
			kd[g_nkd]->lstyle[0] = '1';
			kd[g_nkd]->lstyle[1] = 0;
		}
		kd[g_nkd]->descrip = dp[d]->key_name;
		if (g_get_tex_labels()) {
			kd[g_nkd]->descrip.insert(0, "\\tex{");
			kd[g_nkd]->descrip.append("}");
		}
	}
}

void set_bar_axis_places() {
	for (int bar = 1; bar <= g_nbar; bar++) {
		for (int i = 0; i < br[bar]->ngrp; i++) {
			int to_bar = br[bar]->to[i];
			if (to_bar != 0 && to_bar <= ndata && dp[to_bar] != NULL) {
				axis_struct* axis = br[bar]->horiz ? &xx[2] : &xx[1];
				if (axis->hasNames() && !axis->hasPlaces()) {
					int np = dp[to_bar]->np;
					double* xt = dp[to_bar]->xv;
					if (np == axis->getNbNames()) {
						for (int i = 0; i < np; i++) {
							axis->addPlace(xt[i]);
						}
					}
				}
			}
		}
	}
}

void get_dataset_ranges() {
	// Get ranges from color map
	if (g_colormap != NULL && g_colormap->getData() != NULL) {
		GLERectangle* bounds = g_colormap->getData()->getBounds();
		setrange(bounds->getXMin(), bounds->getYMin(), false);
		setrange(bounds->getXMax(), bounds->getYMax(), false);
	}
	// Set data sets in to/from of bar commands as used
	for (int bar = 1; bar <= g_nbar; bar++) {
		for (int i = 0; i < br[bar]->ngrp; i++) {
			int to_bar = br[bar]->to[i];
			int from_bar = br[bar]->to[i];
			if (to_bar != 0 && to_bar <= ndata && dp[to_bar] != NULL) {
				dp[to_bar]->axisscale = true;
				if (br[bar]->horiz) dp[to_bar]->inverted = true;
				/* Extend range at both sides for a bar data set */
				/* So that bars are visible */
				double delta = bar_get_min_interval(bar, i);
				int np = dp[to_bar]->np;
				double* yt = dp[to_bar]->yv;
				double* xt = dp[to_bar]->xv;
				int* m = dp[to_bar]->miss;
				if (dp[to_bar]->inverted) {
					setrange(yt[0]-delta/2, xt[0], m[0]);
					setrange(yt[np-1]+delta/2, xt[np-1], m[np-1]);
				} else {
					setrange(xt[0]-delta/2, yt[0], m[0]);
					setrange(xt[np-1]+delta/2, yt[np-1], m[np-1]);
				}
			}
			if (from_bar != 0 && from_bar <= ndata && dp[from_bar] != NULL) {
				dp[from_bar]->axisscale = true;
				if (br[bar]->horiz) dp[from_bar]->inverted = true;
			}
		}
	}
	/* If no data set is used, then scale based on all data sets */
	bool has = false;
	for (int dn = 1; dn <= ndata; dn++) {
		if (dp[dn] != NULL && dp[dn]->axisscale) has = true;
	}
	if (!has) {
		for (int dn = 1; dn <= ndata; dn++) {
			if (dp[dn] != NULL) dp[dn]->axisscale = true;
		}
	}
	/* Only scale based on used data sets */
	for (int dn = 1; dn <= ndata; dn++) {
		if (dp[dn] != NULL && dp[dn]->axisscale) {
			int np = dp[dn]->np;
			double* yt = dp[dn]->yv;
			double* xt = dp[dn]->xv;
			int* m = dp[dn]->miss;
			for (int i = 0; i < np; i++) {
				if (dp[dn]->inverted) {
					setrange(yt[i], xt[i], m[i]);
				} else {
					setrange(xt[i], yt[i], m[i]);
				}
			}
		}
	}
}

void set_dataset_minmax_based_on_axis() {
	for (int dn = 1; dn <= ndata; dn++) {
		if (dp[dn] != NULL) {
			int axis = dp[dn]->axis;
			if (axis == GLE_AXIS_X2) {
				axis_struct* ax = &xx[axis];
				if (ax->minset) {
					dp[dn]->xmin = ax->min;
				}
				if (ax->maxset) {
					dp[dn]->xmax = ax->max;
				}
			}
			if (axis == GLE_AXIS_Y2) {
				axis_struct* ax = &xx[axis];
				if (ax->minset) {
					dp[dn]->ymin = ax->min;
				}
				if (ax->maxset) {
					dp[dn]->ymax = ax->max;
				}
			}
		}
	}
}

class GLEColorMapBitmap : public GLEBitmap {
protected:
	GLEZData* m_Data;
	GLEColorMap* m_map;
	double m_ZMin;
	double m_ZMax;
public:
	GLEColorMapBitmap(GLEColorMap* map, GLEZData* data = NULL);
	virtual ~GLEColorMapBitmap();
	GLEBYTE* createColorPalette();
	virtual int readHeader();
	virtual int decode(GLEByteStream* output);
	void plotFunction(GLEPcode& code, int varx, int vary, GLEByteStream* output);
	void plotData(GLEZData* data, GLEByteStream* output);
	inline void setZRange(double zmin, double zmax) { m_ZMin = zmin; m_ZMax = zmax; }
	inline double getZMin() { return m_ZMin; }
	inline double getZMax() { return m_ZMax; }
	inline bool isFunction() { return m_Data == NULL; }
	inline GLEZData* getData() { return m_Data; }
};

GLEColorMapBitmap::GLEColorMapBitmap(GLEColorMap* map, GLEZData* data) : GLEBitmap() {
	m_map = map;
	m_ZMin = 0.0;
	m_ZMax = 0.0;
	m_Data = data;
}

GLEColorMapBitmap::~GLEColorMapBitmap() {
}

int GLEColorMapBitmap::readHeader() {
	m_Width = m_map->getWidth();
	m_Height = m_map->getHeight();
	m_BitsPerComponent = 8;
	if (m_map->isColor() || m_map->hasPalette()) {
		setMode(GLE_BITMAP_RGB);
		setComponents(3);
	} else {
		setMode(GLE_BITMAP_GRAYSCALE);
		setComponents(1);
	}
	return GLE_IMAGE_ERROR_NONE;
}

int fixRange(int v, int min, int max) {
	if (v < min) return min;
	if (v > max) return max;
	return v;
}

void GLEColorMapBitmap::plotData(GLEZData* zdata, GLEByteStream* output) {
	double zvalue;
	double zmin = zdata->getZMin();
	double zmax = zdata->getZMax();
	if (m_map->hasZMin()) zmin = m_map->getZMin();
	if (m_map->hasZMax()) zmax = m_map->getZMax();
	BicubicIpolDoubleMatrix ipd(zdata->getData(), zdata->getNX(), zdata->getNY());
	GLERectangle* bounds = zdata->getBounds();
	int wx1 = (int)floor((m_map->getXMin()-bounds->getXMin())/bounds->getWidth()*(zdata->getNX()-1));
	int wx2 = (int)ceil((m_map->getXMax()-bounds->getXMin())/bounds->getWidth()*(zdata->getNX()-1));
	int wy1 = (int)floor((m_map->getYMin()-bounds->getYMin())/bounds->getHeight()*(zdata->getNY()-1));
	int wy2 = (int)ceil((m_map->getYMax()-bounds->getYMin())/bounds->getHeight()*(zdata->getNY()-1));
	wx1 = fixRange(wx1, 0, zdata->getNX()-1);
	wx2 = fixRange(wx2, 0, zdata->getNX()-1);
	wy1 = fixRange(wy1, 0, zdata->getNY()-1);
	wy2 = fixRange(wy2, 0, zdata->getNY()-1);
	ipd.setWindow(wx1, wy1, wx2, wy2);
	int size = getScanlineSize();
	GLEBYTE* scanline = new GLEBYTE[size];
	int img_hi = getHeight();
	int img_wd = getWidth();
	double scale = zmax - zmin;
	// cout << "zmin = " << zmin << " zmax = " << zmax << endl;
	BicubicIpol ipol(&ipd, img_wd, img_hi);
	if (m_map->isColor()) {
		GLEBYTE* pal = GLEBitmapCreateColorPalette(32761);
		for (int i = img_hi-1; i >= 0; i--) {
			int pos = 0;
			for (int j = 0; j < img_wd; j++) {
				if (m_map->isInverted()) {
					zvalue = (zmax - ipol.ipol(j,i)) / scale;
				} else {
					zvalue = (ipol.ipol(j,i) - zmin) / scale;
				}
				int color = (int)floor(zvalue*32760+0.5);
				if (color > 32760) color = 32760;
				if (color < 0) color = 0;
				scanline[pos++] = pal[color*3];
				scanline[pos++] = pal[color*3+1];
				scanline[pos++] = pal[color*3+2];
			}
			output->send(scanline, size);
			output->endScanLine();
		}
		delete pal;
	} else if (m_map->hasPalette()) {
		GLESub* sub = sub_find(m_map->getPaletteFunction().c_str());
		if (sub == NULL)  {
			stringstream err;
			err << "palette subroutine '" << m_map->getPaletteFunction() << "' not found";
			g_throw_parser_error(err.str());
		} else if (sub->getNbParam() != 1) {
			stringstream err;
			err << "palette subroutine '" << m_map->getPaletteFunction() << "' should take one argument";
			g_throw_parser_error(err.str());
		}
		int otype;
		int nstk = 1;
		char *stk_str[6];
		double stk[6];
		colortyp colvar;
		union {double d; int l[1];} both;
		for (int i = img_hi-1; i >= 0; i--) {
			int pos = 0;
			for (int j = 0; j < img_wd; j++) {
				if (m_map->isInverted()) {
					zvalue = (zmax - ipol.ipol(j,i)) / scale;
				} else {
					zvalue = (ipol.ipol(j,i) - zmin) / scale;
				}
				stk[1] = zvalue;
				sub_call(sub->getIndex(), (double *)&stk, (char **)&stk_str, &nstk, &otype);
				both.d = stk[1];
				colvar.l = both.l[0];
				scanline[pos++] = colvar.b[B_R];
				scanline[pos++] = colvar.b[B_G];
				scanline[pos++] = colvar.b[B_B];
			}
			output->send(scanline, size);
			output->endScanLine();
		}
	} else {
		for (int i = img_hi-1; i >= 0; i--) {
			int pos = 0;
			for (int j = 0; j < img_wd; j++) {
				if (m_map->isInverted()) {
					zvalue = (zmax - ipol.ipol(j,i)) / scale;
				} else {
					zvalue = (ipol.ipol(j,i) - zmin) / scale;
				}
				double grey = floor(zvalue*255+0.5);
				if (grey > 255) grey = 255;
				if (grey < 0) grey = 0;
				scanline[pos++] = (int)grey;
			}
			output->send(scanline, size);
			output->endScanLine();
		}
	}
	setZRange(zmin, zmax);
	delete[] scanline;
}

void GLEColorMapBitmap::plotFunction(GLEPcode& code, int varx, int vary, GLEByteStream* output) {
	int size = getScanlineSize();
	GLEBYTE* scanline = new GLEBYTE[size];
	int img_hi = getHeight();
	int img_wd = getWidth();
	double xmin = m_map->getXMin();
	double ymax = m_map->getYMax();
	double xrange = m_map->getXMax() - xmin;
	double yrange = ymax - m_map->getYMin();
	double zmax = 0.0;
	double zmin = 1.0;
	double scale = 1.0;
	double delta = 0.0;
	double set_zmax = 1.0;
	if (m_map->hasZMin() && m_map->hasZMax()) {
		scale = m_map->getZMax() - m_map->getZMin();
		delta = m_map->getZMin();
		set_zmax = m_map->getZMax();
	}
	if (m_map->isColor()) {
		GLEBYTE* pal = GLEBitmapCreateColorPalette(32761);
		for (int i = 0; i < img_hi; i++) {
			int pos = 0;
			var_set(vary, ymax - (double)i*yrange/img_hi);
			for (int j = 0; j < img_wd; j++) {
				double zvalue;
				var_set(varx, xmin + (double)j*xrange/img_wd);
				eval_pcode(code, &zvalue);
				if (zvalue > zmax) zmax = zvalue;
				if (zvalue < zmin) zmin = zvalue;
				if (m_map->isInverted()) {
					zvalue = scale * (set_zmax - zvalue);
				} else {
					zvalue = scale * (zvalue - delta);
				}
				int color = (int)floor(zvalue*32760+0.5);
				if (color > 32760) color = 32760;
				if (color < 0) color = 0;
				scanline[pos++] = pal[color*3];
				scanline[pos++] = pal[color*3+1];
				scanline[pos++] = pal[color*3+2];
			}
			output->send(scanline, size);
			output->endScanLine();
		}
		delete[] pal;
	} else if (m_map->hasPalette()) {
		GLESub* sub = sub_find(m_map->getPaletteFunction().c_str());
		if (sub == NULL)  {
			stringstream err;
			err << "palette subroutine '" << m_map->getPaletteFunction() << "' not found";
			g_throw_parser_error(err.str());
		} else if (sub->getNbParam() != 1) {
			stringstream err;
			err << "palette subroutine '" << m_map->getPaletteFunction() << "' should take one argument";
			g_throw_parser_error(err.str());
		}
		int otype;
		int nstk = 1;
		char *stk_str[6];
		double stk[6];
		colortyp colvar;
		union {double d; int l[1];} both;
		for (int i = 0; i < img_hi; i++) {
			int pos = 0;
			var_set(vary, ymax - (double)i*yrange/img_hi);
			for (int j = 0; j < img_wd; j++) {
				double zvalue;
				var_set(varx, xmin + (double)j*xrange/img_wd);
				eval_pcode(code, &zvalue);
				if (zvalue > zmax) zmax = zvalue;
				if (zvalue < zmin) zmin = zvalue;
				if (m_map->isInverted()) {
					zvalue = scale * (set_zmax - zvalue);
				} else {
					zvalue = scale * (zvalue - delta);
				}
				stk[1] = zvalue;
				sub_call(sub->getIndex(), (double *)&stk, (char **)&stk_str, &nstk, &otype);
				both.d = stk[1];
				colvar.l = both.l[0];
				scanline[pos++] = colvar.b[B_R];
				scanline[pos++] = colvar.b[B_G];
				scanline[pos++] = colvar.b[B_B];
			}
			output->send(scanline, size);
			output->endScanLine();
		}
	} else {
		for (int i = 0; i < img_hi; i++) {
			int pos = 0;
			var_set(vary, ymax - (double)i*yrange/img_hi);
			for (int j = 0; j < img_wd; j++) {
				double zvalue;
				var_set(varx, xmin + (double)j*xrange/img_wd);
				eval_pcode(code, &zvalue);
				if (zvalue > zmax) zmax = zvalue;
				if (zvalue < zmin) zmin = zvalue;
				if (m_map->isInverted()) {
					zvalue = scale * (set_zmax - zvalue);
				} else {
					zvalue = scale * (zvalue - delta);
				}
				double grey = floor(zvalue*255+0.5);
				if (grey > 255) grey = 255;
				if (grey < 0) grey = 0;
				scanline[pos++] = (int)grey;
			}
			output->send(scanline, size);
			output->endScanLine();
		}
	}
	setZRange(zmin, zmax);
	delete[] scanline;
}

int GLEColorMapBitmap::decode(GLEByteStream* output) {
	if (isFunction()) {
		int varx, vary;
		int vartype = 1;
		var_add_local_submap();
		var_findadd("X", &varx, &vartype);
		var_findadd("Y", &vary, &vartype);
		GLEPcodeList pc_list;
		GLEPcode my_pcode(&pc_list);
		polish((char*)m_map->getFunction().c_str(), my_pcode, &etype);
		plotFunction(my_pcode, varx, vary, output);
		var_remove_local_submap();
	} else {
		plotData(getData(), output);
	}
	var_findadd_set("ZGMIN", getZMin());
	var_findadd_set("ZGMAX", getZMax());
	return GLE_IMAGE_ERROR_NONE;
}

GLEColorMap::GLEColorMap() {
	m_color = false;
	m_wd = 50; m_hi = 50;
	m_xmin = 0.0; m_xmax = 1.0;
	m_ymin = 0.0; m_ymax = 1.0;
	m_zmin = 0.0; m_zmax = 1.0;
	m_has_zmin = false;
	m_has_zmax = false;
	m_invert = false;
	m_haspal = false;
	m_Data = NULL;
}

GLEColorMap::~GLEColorMap() {
	if (m_Data != NULL) delete m_Data;
}

void GLEColorMap::readData() {
	string fname;
	polish_eval_string(getFunction().c_str(), &fname, true);
	if (str_i_ends_with(fname, ".Z")) {
		m_Data = new GLEZData();
		m_Data->read(fname);
	}
}

void GLEColorMap::setPalette(const string& pal) {
	m_palette = pal;
	m_haspal = true;
}

void GLEColorMap::setXRange(double min, double max) {
	m_xmin = min; m_xmax = max;
}

void GLEColorMap::setYRange(double min, double max) {
	m_ymin = min; m_ymax = max;
}

void GLEColorMap::setZMin(double val) {
	m_has_zmin = true;
	m_zmin = val;
}

void GLEColorMap::setZMax(double val) {
	m_has_zmax = true;
	m_zmax = val;
}

void GLEColorMap::draw(double x0, double y0, double wd, double hi) {
	GLEZData* zdata = getData();
	if (zdata != NULL) {
		/* figure out position of bitmap */
		GLERectangle* bounds = zdata->getBounds();
		double xrange = getXMax() - getXMin();
		double yrange = getYMax() - getYMin();
		double xmin = (bounds->getXMin()-getXMin())/xrange*wd;
		if (xmin > wd) return;
		if (xmin < 0) xmin = 0;
		double ymin = (bounds->getYMin()-getYMin())/yrange*hi;
		if (ymin > hi) return;
		if (ymin < 0) ymin = 0;
		double xmax = (bounds->getXMax()-getXMin())/xrange*wd;
		if (xmax < 0) return;
		if (xmax > wd) xmax = wd;
		double ymax = (bounds->getYMax()-getYMin())/yrange*hi;
		if (ymax < 0) return;
		if (ymax > hi) ymax = hi;
		/* draw bitmap */
		g_move(x0 + xmin, y0 + ymin);
		GLEColorMapBitmap bitmap(this, zdata);
		g_bitmap(&bitmap, xmax-xmin, ymax-ymin, BITMAP_TYPE_USER);
	} else {
		g_move(x0, y0);
		GLEColorMapBitmap bitmap(this);
		g_bitmap(&bitmap, wd, hi, BITMAP_TYPE_USER);
	}
}
